#include <omp.h>

#define Nmax (1<<24)

/**
 * Generic synthetic kernels.
 *
 * GPU-kernel Kernel_a_b has a ingoing and b outgoing operands.
 * CPU-kernel called CPU_Kernel_a_b ( numElements, ... )
 */
__global__ void
Kernel_1_0 (int numElements, float *A )
{
   int i = blockDim.x * blockIdx.x + threadIdx.x;
   int t = 1;  
   while (t < numElements) { 
    if (i < numElements/t) {
        A[i] += A[i+t];
    }
    t = t * 2; 
   }
}

void CPUKernel_1_0 (int numElements, const float *A )
{
 int i;
 float B=0.0; 
 for (i=0; i < numElements; i++) {
     B +=  A[i] * A[i];
 } 
}

__global__ void
Kernel_2_0 (int numElements, const float *A, const float *B )
{
   int i = blockDim.x * blockIdx.x + threadIdx.x;
   float reg; 
   int t = 2;  
   while (t < numElements) { 
    if (i < numElements/t) {
        reg += A[2*i] * B[2*i+1];
    }
    t = t * 2; 
   }
}

void CPUKernel_2_0 (int numElements, const float *A, const float *B )
{
 int i;
 float s=0.0; 
 for (i=0; i < numElements; i++) {
     s +=  A[i] * B[i];
 }
}

__global__ void
Kernel_1_1 (int numElements, const float *A, float *B )
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < numElements) {
        B[i] = A[i] * A[i];
    }
}

void CPUKernel_1_1 (int numElements, const float *A, float *B )
{
 int i;
 for (i=0; i < numElements; i++) {
     B[i] =  A[i] * A[i];
 } 
}

__global__ void
Kernel_1_2 (int numElements, const float *A, float *B, float *C )
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < numElements) {
        B[i] = A[i] * A[i];
        C[i] = B[i] + A[i];
    }
}

void CPUKernel_1_2 (int numElements, const float *A, float *B, float *C )
{
 int i;
 for (i=0; i < numElements; i++) {
     B[i] =  A[i] * A[i];
     C[i] =  B[i] + A[i];
 } 
}

__global__ void
Kernel_2_1 (int numElements, const float *A, const float *B, float *C)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < numElements) {
        C[i] = A[i] + B[i];
    }
}

void CPUKernel_2_1 (int numElements, const float *A, const float *B, float *C )
{
 int i;
 for (i=0; i < numElements; i++) {
     C[i] = A[i] + B[i];
 } 
}

__global__ void
Kernel_2_2 (int numElements, const float *A, const float *B, float *C, float *D)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < numElements) {
        C[i] = A[i] + B[i];
        D[i] = A[i] - B[i];
    }
}

void CPUKernel_2_2 (int numElements, const float *A, const float *B, float *C, float *D )
{
 int i;
 for (i=0; i < numElements; i++) {
     C[i] = A[i] + B[i];
     D[i] = A[i] + B[i];
 } 
}

__global__ void
Kernel_3_2 (int numElements, const float *A, const float *B, const float *C, float *D, float *E)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < numElements) {
        D[i] = A[i] + B[i]*C[i];
        E[i] = C[i] - B[i];
    }
}

void CPUKernel_3_2 (int numElements, const float *A, const float *B, const float *C, float *D, float *E )
{
 int i;
 for (i=0; i < numElements; i++) {
     D[i] = A[i] + B[i]*C[i];
     E[i] = C[i] - B[i];
 }
}



/* Application-specific kernels.
 * 
 * For kernel NAME:
 * GPU-kernel NAME_a_b has a ingoing and b outgoing operands.
 * CPU-kernel CPUNAME_Kernel_a_b ( numElements, ... )
 */ 

// GEMM_2_1: GEMM with no offset:

__global__ void
GEMM_2_1 (int numElements, const float *A, const float *B, float *C )
{
 int i, j, k, ij;
 int numRows = (int)sqrtf(numElements);
 ij = blockDim.x * blockIdx.x + threadIdx.x;
 if (ij < numElements) {
    i = ij / numRows;
    j = ij % numRows;
    for (k=0; k < numRows; k++) {
       C[ij] += A[i*numRows+k] * B[k*numRows+j]; 
    }
 }
}

void CPUGEMM_2_1 (int numElements, const float *A, const float *B, float *C )
{
 int i, j, k;
 int numRows = (int)sqrt(numElements);
 for (i=0; i < numRows; i++) {
    for (j=0; j < numRows; j++) {
       for (k=0; k < numRows; k++) {
          C[i*numRows+j] += A[i*numRows+k] * B[k*numRows+j]; 
       }
    }
 }
}

// GEMM_3_1: GEMM with offset  C_new = C_old + A x B
__global__ void
GEMM_3_1 (int numElements, const float *C_old, const float *A, const float *B, float *C )
{
 int i, j, k, ij;
 int numRows = (int)sqrtf(numElements);
 ij = blockDim.x * blockIdx.x + threadIdx.x;
 if (ij < numElements) {
    C[ij] = C_old[ij]; 
    i = ij / numRows;
    j = ij % numRows;
    for (k=0; k < numRows; k++) {
       C[ij] += A[i*numRows+k] * B[k*numRows+j]; 
    }
 }
}

void CPUGEMM_3_1 (int numElements, const float *C_old, const float *A, const float *B, float *C )
{
 int i, j, k;
 int numRows = (int)sqrt(numElements);
#pragma omp parallel for (schedule="STATIC")
 for (i=0; i < numRows; i++) {
    for (j=0; j < numRows; j++) {
       C[i*numRows+j] = C_old[i*numRows+j]; 
       for (k=0; k < numRows; k++) {
          C[i*numRows+j] += A[i*numRows+k] * B[k*numRows+j]; 
       }
    }
 }
}


/* SYRK: performs one of the symmetric rank k operations
 *    C := alpha*A*A**T + beta*C,
 * or
 *    C := alpha*A**T*A + beta*C,
 * where  alpha and beta  are scalars, C is an  n by n  symmetric matrix
 * and  A  is an  n by k  matrix in the first case and a  k by n  matrix
 * in the second case. 
 * Strongly simplified here (e.g. k=n, alpha=1, beta=1)
 */

__global__ void
SYRK_2_1 (int numElements, const float *A, const float *Cin, float *C )
{
 int i, j, k, ij;
 int numRows = (int) sqrtf(numElements);
 ij = blockDim.x * blockIdx.x + threadIdx.x;
 if (ij < numElements) {
    i = ij / numRows;
    j = ij % numRows;
    int K=1; // shortcut as on CPU
    C[ij] = Cin[ij];
    for (k=0; k < K; k++) {
          C[ij] += A[i*numRows+k] * A[k*numRows+j]; 
    }
 }
}

void CPUSYRK_2_1 (int numElements, const float *A, const float *Cin, float *C )
{
 int i, j, l;
 /*from netlib:
   DO 180 j = 1,n
                   IF (beta.EQ.zero) THEN
                       DO 140 i = j,n
                           c(i,j) = zero
   140                 CONTINUE
                   ELSE IF (beta.NE.one) THEN
                       DO 150 i = j,n
                           c(i,j) = beta*c(i,j)
   150                 CONTINUE
                   END IF
                   DO 170 l = 1,k
                       IF (a(j,l).NE.zero) THEN
                           temp = alpha*a(j,l)
                           DO 160 i = j,n
                               c(i,j) = c(i,j) + temp*a(i,l)
   160                     CONTINUE
                       END IF
   170             CONTINUE
   180 CONTINUE
   */
 // We consider only the case k==1 for simplicity:
 for (j=0; j < numElements; j++) 
       C[j] = Cin[j];	

 int numRows = (int) sqrt(numElements);
 int k = 1; // shortcut
 for (j=0; j < numRows; j++) {
    for (l=0; l < k; l++) {
       float temp = A[l*numRows+j];
       if (temp != 0.0) {
        for (i=j; i < numRows; i++) {
          C[j*numRows+i] += temp * A[l*numRows+i]; 
        }
       }
    }
 }
}

/* Special case of SYRK for one input operand as occurs in Cholesky */

__global__ void
SYRK_1_1 (int numElements, const float *A, float *C )
{
 int i, j, k, ij;
 int numRows = (int)sqrtf(numElements);
 ij = blockDim.x * blockIdx.x + threadIdx.x;
 if (ij < numElements) {
    i = ij / numRows;
    j = ij % numRows;
    C[ij] = 0.0;
    for (k=0; k < numRows; k++) {
          C[ij] += A[i*numRows+k] * A[k*numRows+j]; 
    }
 }
}

void CPUSYRK_1_1 (int numElements, const float *A, float *C )
{
 int i, j, l;
 int numRows = (int)sqrt(numElements);
 for (j=0; j < numElements; j++) 
       C[j] = 0.0;	
 int k = 1; // shortcut
 for (j=0; j < numRows; j++) {
    for (l=0; l < k; l++) {
       float temp = A[l*numRows+j];
       if (temp != 0.0) {
        for (i=j; i < numRows; i++) {
          C[j*numRows+i] += temp * A[l*numRows+i]; 
        }
       }
    }
 }
}


/* TRSM: solve upper or lower tridiagonal system Ax=X */

__global__ void
TRSM_2_1 (int numElements, const float *A, const float *X, float *x )
{
 int i, j;
 i = blockDim.x * blockIdx.x + threadIdx.x;
 x[i] = X[i];  // need SSA property, not in-place
 if (i < numElements) {
    if (x[i]!=0.0) {
      float temp = x[i] / A[i*numElements+i];
      for (j=i+1; j < numElements; j++) {
         x[j] -= temp*A[i*numElements+j]; 
      } 
    }
 }
}

void CPUTRSM_2_1 (int numElements, const float *A, const float *X, float *x )
{
 int i, j;

 // Simple case only, nontransposed A
 //                  jx = kx
 //                  DO 80 j = 1,n
 //                      IF (x(jx).NE.zero) THEN
 //                          IF (nounit) x(jx) = x(jx)/a(j,j)
 //                          temp = x(jx)
 //                          ix = jx
 //                          DO 70 i = j + 1,n
 //                              ix = ix + incx
 //                              x(ix) = x(ix) - temp*a(i,j)
 //   70                     CONTINUE
 //                      END IF
 //                      jx = jx + incx
 //   80             CONTINUE

 for (i=0; i < numElements; i++) 
    x[i] = X[i];  // need SSA property, not in-place
 for (i=0; i < numElements; i++) {
    if (x[i]!=0.0) {
      float temp = x[i] / A[i*numElements+i];
      for (j=i+1; j < numElements; j++) {
         x[j] -= temp*A[i*numElements+j]; 
      } 
    }
 }
}

/* Special case of TRSM with one operand (the diagonal of A) */

__global__ void
TRSM_1_1 (int numElements, const float *A, float *x )
{
 int i, j;
 i = blockDim.x * blockIdx.x + threadIdx.x;
 x[i] = A[i*numElements+i];
 if (i < numElements) {
    if (x[i]!=0.0) {
      float temp = x[i] / A[i*numElements+i];
      for (j=i+1; j < numElements; j++) {
         x[j] -= temp*A[i*numElements+j]; 
      }
    }
 }
}

void CPUTRSM_1_1 (int numElements, const float *A, float *x )
{
 int i, j;

 // Simple case only, nontransposed A
 //                  jx = kx
 //                  DO 80 j = 1,n
 //                      IF (x(jx).NE.zero) THEN
 //                          IF (nounit) x(jx) = x(jx)/a(j,j)
 //                          temp = x(jx)
 //                          ix = jx
 //                          DO 70 i = j + 1,n
 //                              ix = ix + incx
 //                              x(ix) = x(ix) - temp*a(i,j)
 //   70                     CONTINUE
 //                      END IF
 //                      jx = jx + incx
 //   80             CONTINUE

 for (i=0; i < numElements; i++) 
    x[i] = A[i*numElements+i];
 for (i=0; i < numElements; i++) {
    if (x[i]!=0.0) {
      float temp = x[i] / A[i*numElements+i];
      for (j=i+1; j < numElements; j++) {
         x[j] -= temp*A[i*numElements+j]; 
      } 
    }
 }
}


/* POTRF - this must be the call to POTF2 in Cholesky, see netlib: *
 * SPOTF2 computes the Cholesky factorization of a real symmetric
 * positive definite matrix A.
 * The factorization has the form
 *   A = U**T * U ,  if UPLO = 'U', or
 *   A = L  * L**T,  if UPLO = 'L',
 * where U is an upper triangular matrix and L is lower triangular.
 * This is the unblocked version of the algorithm, calling Level 2 BLAS. 
 * Much simplified here, i.e. only for Upper triangular matrix */

void CPUPOTRF_1_1 (int numElements, const float *A, float *Aout )
{
 int i, j;
 int numRows = (int)sqrt(numElements);

 /* From netlib code: Compute the Cholesky factorization A = U**T *U.
   *
            DO 10 j = 1, n
   *
   *           Compute U(J,J) and test for non-positive-definiteness.
   *
               ajj = a( j, j ) - sdot( j-1, a( 1, j ), 1, a( 1, j ), 1 )
               IF( ajj.LE.zero.OR.sisnan( ajj ) ) THEN
                  a( j, j ) = ajj
                  go to 30
               END IF
               ajj = sqrt( ajj )
               a( j, j ) = ajj
   *
   *           Compute elements J+1:N of row J.
   *
               IF( j.LT.n ) THEN
                  CALL sgemv( 'Transpose', j-1, n-j, -one, a( 1, j+1 ),
        $                     lda, a( 1, j ), 1, one, a( j, j+1 ), lda )
                  CALL sscal( n-j, one / ajj, a( j, j+1 ), lda )
               END IF
      10    continue

      The computation below is not correct but performs the same work:
  */

 for (j=0; j<numRows; j++) {
    float ajj = A[j*numRows+j];
    if (ajj > 0.0) {
     for (i=0; i<j; i++) { // SDOT: 
       ajj -= A[j*numRows+i] * A[j*numRows+i];
     }
     Aout[j*numRows+i] = ajj;
     for (i=j+1; i<numRows; i++) {
       Aout[j*numRows+i] -= Aout[j*numRows+0] * Aout[i*numRows+j] / ajj;
     }
    } else {
     Aout[j*numRows+j] = ajj;
     // skip further treatment of special case 
    }
 }
}


/* MADD: */

__global__ void
MADD_2_1 (int numElements, const float *A, const float *B, float *C)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < numElements) {
        C[i] = A[i] + B[i];
    }
}

void CPUMADD_2_1 (int numElements, const float *A, const float *B, float *C )
{
 int i;
 for (i=0; i < numElements; i++) {
     C[i] = A[i] + B[i];
 }
}

/* VADD: */

__global__ void
VADD_2_1 (int numElements, const float *A, const float *B, float *C)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < numElements) {
        C[i] = A[i] + B[i];
    }
}

void CPUVADD_2_1 (int numElements, const float *A, const float *B, float *C )
{
 int i;
 for (i=0; i < numElements; i++) {
     C[i] = A[i] + B[i];
 }
}


/* VMUL: */

__global__ void
VMUL_2_1 (int numElements, const float *A, const float *B, float *C)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < numElements) {
        C[i] = A[i] * B[i];
    }
}

void CPUVMUL_2_1 (int numElements, const float *A, const float *B, float *C )
{
 int i;
 for (i=0; i < numElements; i++) {
     C[i] = A[i] * B[i];
 }
}


/* VSCALE: */

__global__ void
VSCALE_1_1 (int numElements, const float *A, float *C)
{
 int i = blockDim.x * blockIdx.x + threadIdx.x;
 float factor = 3.14; // dummy
 if (i < numElements) {
        C[i] = A[i] * factor;
 }
}

void CPUVSCALE_1_1 (int numElements, const float *A, float *C )
{
 int i;
 float factor = 3.14; // dummy
 for (i=0; i < numElements; i++) {
     C[i] = A[i] * factor;
 }
}


/* SAXPY: */

__global__ void
SAXPY_2_1 (int numElements, const float *A, const float *B, float *C)
{
 int i = blockDim.x * blockIdx.x + threadIdx.x;
 float factor = 3.14; // dummy
 if (i < numElements) {
        C[i] = A[i] * factor + B[i];
 }
}

void CPUSAXPY_2_1 (int numElements, const float *A, const float *B, float *C )
{
 int i;
 float factor = 3.14; // dummy
 for (i=0; i < numElements; i++) {
     C[i] = A[i] * factor + B[i];
 }
}

__global__ void
SAXPY_1_1 (int numElements, const float *A, float *C)
{
 int i = blockDim.x * blockIdx.x + threadIdx.x;
 float factor = 3.14; // dummy
 if (i < numElements) {
        C[i] = A[i] * factor + A[i];
 }
}

void CPUSAXPY_1_1 (int numElements, const float *A, float *C )
{
 int i;
 float factor = 3.14; // dummy
 for (i=0; i < numElements; i++) {
     C[i] = A[i] * factor + A[i];
 }
}

/* SDOT: */

__global__ void
SDOT_2_0 (int numElements, float *A, const float *B)
{
 int i = blockDim.x * blockIdx.x + threadIdx.x;
 if (i < numElements) {
   float reg = A[i] * B[i];
   // skip sync for simplicity 
   A[i] = reg;
   int t = 2;
   while (t < numElements) { 
    if (i < numElements/t) {
        reg += A[i+t];
    }
    t = t * 2; 
    // skip sync for simplicity 
   }
 }
}

void CPUSDOT_2_0 (int numElements, const float *A, const float *B )
{
 int i;
 float sum;
 for (i=0; i < numElements; i++) {
     sum += A[i] * B[i];
 }
}

__global__ void
SDOT_1_0 (int numElements, float *A)
{
 int i = blockDim.x * blockIdx.x + threadIdx.x;
 if (i < numElements) {
   float reg = A[i] = A[i] * A[i];
   // skip sync for simplicity 
   int t = 2;
   while (t < numElements) { 
    if (i < numElements/t) {
        reg += A[i+t];
    }
    t = t * 2; 
   }
 }
}

void CPUSDOT_1_0 (int numElements, const float *A )
{
 int i;
 float sum;
 for (i=0; i < numElements; i++) {
     sum += A[i] * A[i];
 }
}


/* SPMV_1_1  sparse matrix-vector multiply - buggy, code generation should go to CUSPARSE instead */

// sparse matrix data structure not modeled
int firstinrow[Nmax];
int col[Nmax];

void CPUSPMV_1_1 (int numRows, const float *A, float *x )
{
 int i, j;
 static int initialized = 0;
 float sum;
 if (!initialized) { // do this only once:
  for (i=0; i <= numRows; i++) {
    // dummy solution for now - assume 3 nonzeros per row:
    firstinrow[i]=3*i; 
  }
  for (i=0; i < 3*numRows; i++) {
    col[i]= 3*i;
  }
  initialized = 1;
 }
 for (i=0; i < numRows; i++) {
    x[i] = 0.0;
    for (j=firstinrow[i]; j < firstinrow[i+1]; j++)
       x[i] += A[j] * x[col[j]];
 }
}
