#include <stdio.h>
#include <sys/time.h>

// For the CUDA runtime routines (prefixed with "cuda_")
#include <cuda_runtime.h>

//#include <helper_cuda.h>

// repetitions to reduce time variation:
#define NITER 200

#ifndef LOGVLENGTH 
#define LOGVLENGTH 10
#endif

#define NUMELEMENTS 1<<LOGVLENGTH

// activate to measure cudaMalloc time in each iteration of NITER
// #define INCLUDECUDAMALLOC   ---- now set externally in Makefile

/**
 * Computes the vector addition of A and B into C. The 3 vectors have the same
 * number of elements numElements.
 */
__global__ void
vectorAdd(const float *A, const float *B, float *C, int numElements)
{
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if (i < numElements) 
        C[i] = A[i] + B[i];
}

/** Host main routine:
 */
int main(void)
{
    struct timeval starttime, midtime, midtime2, endtime, nkftime, nkfstarttime;
    int totaltime; 
    int fusedtime; 
    int nokfusedtime; 
    float *d_A, *d_B, *d_C;
    int iter;

    // Error code to check return values for CUDA calls
    // cudaError_t err = cudaSuccess;

    // do the expensive initial dummy call to cudaMalloc:
    float *dummy;
    cudaMalloc((void **)&dummy, 256);

    // Print the vector length to be used, and compute its size
    int numElements = NUMELEMENTS;
    int threadsPerBlock = 256;
    int blocksPerGrid = (numElements + threadsPerBlock - 1) / threadsPerBlock;
    size_t size = numElements * sizeof(float);
    printf("[Vector addition of %d elements]\n", numElements);
#ifdef INCLUDECUDAMALLOC
    printf("[cudaMalloc calls are included]\n");
#endif

    // Allocate the host input vectors:
    // float *h_A = (float *)malloc(size);
    // float *h_B = (float *)malloc(size);
    // float *h_C = (float *)malloc(size);
    //testweise in page-locked (pinned) memory:
    float *h_A; cudaMallocHost((void **)&h_A, size);
    float *h_B; cudaMallocHost((void **)&h_B, size);
    float *h_C; cudaMallocHost((void **)&h_C, size);

    // Initialize the host input vectors
    for (int i = 0; i < numElements; ++i)
    {
        h_A[i] = rand()/(float)RAND_MAX;
        h_B[i] = rand()/(float)RAND_MAX;
    }

    // Allocate device input vectors:
#ifndef INCLUDECUDAMALLOC
    cudaMalloc((void **)&d_A, size);
    cudaMalloc((void **)&d_B, size);
    cudaMalloc((void **)&d_C, size);
#endif

 cudaDeviceSynchronize();
 gettimeofday( &starttime, NULL);  // take the time

 for (iter=0; iter<NITER; iter++) 
 {
    // Allocate device input vectors:
#ifdef INCLUDECUDAMALLOC
    cudaMalloc((void **)&d_A, size);
    cudaMalloc((void **)&d_B, size);
    cudaMalloc((void **)&d_C, size);
#endif

    // Copy the host input vectors A and B in host memory to the device input vectors in
    // device memory
    // printf("Copy input data from the host memory to the CUDA device\n");
    cudaMemcpy(d_A, h_A, size, cudaMemcpyHostToDevice);
    cudaMemcpy(d_B, h_B, size, cudaMemcpyHostToDevice);

    // Launch the Vector Add CUDA Kernel
    vectorAdd<<<blocksPerGrid, threadsPerBlock>>>(d_A, d_B, d_C, numElements);
    //cudaGetLastError();
    cudaMemcpy(h_C, d_C, size, cudaMemcpyDeviceToHost);
    // Free device global memory
#ifdef INCLUDECUDAMALLOC
    cudaFree(d_A);
    cudaFree(d_B);
    cudaFree(d_C);
#endif
 }
    cudaDeviceSynchronize();
    gettimeofday( &midtime, NULL);  // take the time

#ifndef INCLUDECUDAMALLOC
    cudaFree(d_A);
    cudaFree(d_B);
    cudaFree(d_C);
#endif

    totaltime = (midtime.tv_sec - starttime.tv_sec) * 1000000
              + (midtime.tv_usec - starttime.tv_usec);
    printf(" Only Kernel-fused time (%d iter.): %d us\n", NITER, totaltime ); 

    // Verify that the result vector is correct
    for (int i = 0; i < numElements; ++i) {
        if (fabs(h_A[i] + h_B[i] - h_C[i]) > 1e-5) {
            fprintf(stderr, "Result verification failed at element %d!\n", i);
            exit(EXIT_FAILURE);
        }
    }
    // free(h_A);
    // free(h_B);
    // free(h_C);
    // testweise pagelocked(pinned) memory:
    cudaFreeHost(h_A);
    cudaFreeHost(h_B);
    cudaFreeHost(h_C);


    // new test, now fused:  ---------------

    //testweise in page-locked (pinned) memory:
    cudaMallocHost((void **)&h_A, 3*size);
    // h_A = (float *)malloc( 3*size );
    h_B = h_A + numElements;
    h_C = h_B + numElements;

    // Initialize the host input vectors
    for (int i = 0; i < numElements; ++i)
    {
        h_A[i] = rand()/(float)RAND_MAX;
        h_B[i] = rand()/(float)RAND_MAX;
    }

#ifndef INCLUDECUDAMALLOC
    cudaMalloc((void **)&d_A, 3*size);
    d_B = d_A + numElements; // instead of cudaMalloc((void **)&d_B, size);
    //cudaMalloc((void **)&d_C, size); // no allocation fusion here
    d_C = d_B + numElements; // instead of cudaMalloc((void **)&d_B, size);
#endif
 
 cudaDeviceSynchronize();
 gettimeofday( &midtime2, NULL);  // take the time

 for (iter=0; iter<NITER; iter++)
 { 
    // Allocate device input vectors:
#ifdef INCLUDECUDAMALLOC
    cudaMalloc((void **)&d_A, 3*size);
    d_B = d_A + numElements; // instead of cudaMalloc((void **)&d_B, size);
    //cudaMalloc((void **)&d_C, size); // no allocation fusion here
    d_C = d_B + numElements; // instead of cudaMalloc((void **)&d_B, size);
#endif

    cudaMemcpy(d_A, h_A, 2*size, cudaMemcpyHostToDevice);
    // cudaMemcpy(d_B, h_B, size, cudaMemcpyHostToDevice);

    // Launch the Vector Add CUDA Kernel
    vectorAdd<<<blocksPerGrid, threadsPerBlock>>>(d_A, d_B, d_C, numElements);
    //cudaGetLastError();
    cudaMemcpy(h_C, d_C, size, cudaMemcpyDeviceToHost);
    // Free device global memory
#ifdef INCLUDECUDAMALLOC
    cudaFree(d_A);
    //cudaFree(d_C);
#endif
 }
    cudaDeviceSynchronize();
    gettimeofday( &endtime, NULL);  // take the time
#ifndef INCLUDECUDAMALLOC
    cudaFree(d_A);
    //cudaFree(d_C);
#endif

    fusedtime = (endtime.tv_sec - midtime2.tv_sec) * 1000000
              + (endtime.tv_usec - midtime2.tv_usec);
    printf(" Transfer(+Kernel)-fused Time (%d iter.): %d us\n", NITER, fusedtime ); 
    printf(" Transfer fusion speedup per iter.: %d us = %2.1f%% of totaltime %d us\n",
            (int) ((float)(totaltime - fusedtime)/(float)NITER),
            (100.0*(float)(totaltime - fusedtime))/(float)totaltime,
            totaltime / NITER );

    // Verify that the result vector is correct
    for (int i = 0; i < numElements; ++i) {
        if (fabs(h_A[i] + h_B[i] - h_C[i]) > 1e-5) {
            fprintf(stderr, "Result verification failed at element %d!\n", i);
            exit(EXIT_FAILURE);
        }
    }

    // Free host memory
    // free(h_A);
    // testwise page-locked (pinned) memory:
    cudaFreeHost(h_A);


    // new test, now message-fused but non-kernelfused:  ---------------

    // twestwise in pagelocked (pinned) memory:
    cudaMallocHost((void **)&h_A, 3*size);
    // h_A = (float *)malloc( 3*size );
    h_B = h_A + numElements;
    h_C = h_B + numElements;

    // Initialize the host input vectors
    for (int i = 0; i < numElements; ++i)
    {
        h_A[i] = rand()/(float)RAND_MAX;
        h_B[i] = rand()/(float)RAND_MAX;
    }

#ifndef INCLUDECUDAMALLOC
    cudaMalloc((void **)&d_A, 3*size);
    d_B = d_A + numElements; // instead of cudaMalloc((void **)&d_B, size);
    d_C = d_B + numElements; // instead of cudaMalloc((void **)&d_B, size);
    //cudaMalloc((void **)&d_C, size);
#endif

 cudaDeviceSynchronize();
 gettimeofday( &nkfstarttime, NULL);  // take the time

 for (iter=0; iter<NITER; iter++)
 { 
    // Allocate device input vectors:
#ifdef INCLUDECUDAMALLOC
    cudaMalloc((void **)&d_A, 3*size);
    d_B = d_A + numElements; // cudaMalloc((void **)&d_B, size);
    // Test if different allocation sizes cause the anomaly for small vectors:
    //if (numElements <=65536) cudaMalloc((void **)&d_C, 2*size);
    // else   // no, no difference, the anomaly remains.
    //cudaMalloc((void **)&d_C, size);
    d_C = d_B + numElements; // instead of cudaMalloc((void **)&d_B, size);
#endif

    cudaMemcpy(d_A, h_A, 2*size, cudaMemcpyHostToDevice);
    // cudaMemcpy(d_B, h_B, size, cudaMemcpyHostToDevice);

    // Launch the Vector Add CUDA Kernel, NOT FUSED:
    vectorAdd<<<blocksPerGrid, threadsPerBlock>>>(d_A, d_B, d_C, numElements/2);
    vectorAdd<<<blocksPerGrid, threadsPerBlock>>>(d_A+numElements/2, d_B+numElements/2, d_C+numElements/2, numElements/2);
    //cudaGetLastError();
    cudaDeviceSynchronize();
    cudaMemcpy(h_C, d_C, size, cudaMemcpyDeviceToHost);
    // Free device global memory
#ifdef INCLUDECUDAMALLOC
    cudaFree(d_A);
    //cudaFree(d_C);
#endif
 }
 cudaDeviceSynchronize();
 gettimeofday( &nkftime, NULL);  // take the time
#ifndef INCLUDECUDAMALLOC
    cudaFree(d_A);
    //cudaFree(d_C);
#endif

    nokfusedtime = (nkftime.tv_sec - nkfstarttime.tv_sec) * 1000000
              + (nkftime.tv_usec - nkfstarttime.tv_usec);
    printf(" Transfer- but not Kernel-fused time (%d iter.): %d us\n", NITER, nokfusedtime ); 
    printf(" Kernel fusion speedup: %d us = %2.1f%% of all-fused totaltime %d us\n",
            (int) ((float)(nokfusedtime - fusedtime)/(float)NITER),
            (100.0*(float)(nokfusedtime - fusedtime))/(float)nokfusedtime,
            nokfusedtime / NITER );

    // Verify that the result vector is correct
    for (int i = 0; i < numElements; ++i) {
        if (fabs(h_A[i] + h_B[i] - h_C[i]) > 1e-5) {
            fprintf(stderr, "Result verification failed at element %d!\n", i);
            exit(EXIT_FAILURE);
        }
    }

    // Free host memory
    //free(h_A);
    cudaFreeHost(h_A);

    printf("Done\n");
    return 0;
}
