#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <sys/time.h>
#include "glob.h"

// N number of kernel calls, #defined in command line (default in glob.h)

// INIT_K = number of (pre-placed) input vectors to use.
//        Each kernel's outputs are written to fresh vectors.
// Hence, MAX_K must be >= INIT_K + sum_i=0...N-1 (outdeg[i]).
// The vectors are numbered 0...k-1 with k <= MAX_K

// for Kernel N, need up to #live-on-exit many in[N][.] entries:
#define MAX_MAX_INDEG max(MAX_INDEG,N)

// the kernel-vector graph:
char *name[N+1]; // kernel function name, new 190302
int indeg[N+1]; // last kernel is artificial (marking vectors live on exit)
int outdeg[N];
int in[N+1][MAX_MAX_INDEG]; // refer to vector numbers in 0...K-1
                        // artificial kernel N: live-on-exit
int out[N][MAX_OUTDEG]; // refer to vector numbers in 0...K-1

// the vector-kernel graph (dual graph):
int prod[MAX_K+1][N]; // producers of each vector (if any)
int nprod[MAX_K+1];   
int cons[MAX_K+1][N]; // consumers of each vector (if any)
int ncons[MAX_K+1];   

int k = INIT_K; // actually used number of vectors, INIT_K <= k <= MAX_K
int kk;         // used number of colors, kk <= k
int location[MAX_K+1]; // 1=CPU, 2=GPU, 3=both, 0 = not allocated
int vec[MAX_K+1][MAX_K+1]; // maps kk colors, each to set of vectors having it
int nvec[MAX_K+1]; // cardinality of each color set 

int sched[N];  // permutation of 0...N-1, schedule of operations.
               // initially, sched[i] == i. Live ranges dep. on sched.
int live[N+1][MAX_K+1]; // 1 iff vector j live at kernel i.
                      // N+1: live on exit (artificial kernel).

float aff[MAX_K+1][MAX_K+1]; // k x k weight matrix giving the benefit of
                       // having vector j directly follow i in memory.
int up_earliest[MAX_K+1]; // earliest time (kernel number) where vector v
                        // can be uploaded to GPU. 0 for v in 0..INIT_K-1.
int up_latest[MAX_K+1]; // latest time (kernel number) where vector v
                      // can be uploaded to GPU. 0 for v in 0..INIT_K-1.
int down_earliest[MAX_K+1]; // earliest time (kernel number) where vector v
                          // can be downloaded from GPU.
int down_latest[MAX_K+1]; // latest time (kernel number) where vector v
                        // can be downloaded from GPU.
#define EPSILON_UP 0.5    /* affinity bonus for uploading siblings together */
#define EPSILON_DOWN 0.5  /* affinity bonus for uploading siblings together */
int up_possible[N][MAX_K+1]; // [i][j] == 1 iff i in up_earliest[j]...up_latest[j]
int down_possible[N+1][MAX_K+1]; // [i][j] == 1 iff i in down_earliest[j]...down_latest[j]
int uploads_from[N];  // position in memory of first bulk upload operand before i,
                      // or -1 if not uploaded
int uploads_until[N]; // position in memory of last bulk upload operand before i
int downloads_from[N+1];  // position in memory of first bulk download operand before i
                        // or -1 if not downloaded
int downloads_until[N+1]; // position in memory of last bulk download operand before i

int calledmalloc[MAX_K+1];  // memory-allocated vectors (permuted)

int color[MAX_K+1];  // memory-allocated vectors (permuted)

int device[N+1];  // 0=CPU, 1=GPU. Kernel N is artificial (for live on exit)

#ifndef Nattempts
#define Nattempts 4  /* default number of attempts for Hamiltonian heuristic to pick the best one */
#endif

float hamilton[Nattempts][MAX_K+1][MAX_K+1];
int hamlength[Nattempts]; // the lenghts of Hamiltonian paths for all attempts
float weightedhamlength[Nattempts]; // the lenghts of Hamiltonian paths for all attempts
int bestweightedhamilton; // will finally index the best one
int bestweightedhamlength;
//int besthamlength;
int visited[MAX_K+1]; // needed by DFS
int position[MAX_K+1]; // the final memory mapping of the vectors
int rev_position[MAX_K+1]; // the reverse final memory mapping of the vectors

// Result:
int savings;

// Timing:
struct timeval starttime, endtime,
               /* for subtimings: */
               elstarttime, affstarttime, hamstarttime, emitnaivestarttime, emitstarttime, printstarttime; 

// min
#define min(u,v) ((u)<=(v)?(u):(v))

/* overlap: return 1 iff intervals [e1:l1] and [e2:l2] overlap
 */
int overlap( int e1, int l1, int e2, int l2 )
{
 assert(e1<=l1); // sanity check
 assert(e2<=l2); // sanity check
 if (e1<=e2) return ((e2<=l1)?1:0);
 else        return ((e1<=l2)?1:0);
}

/* lett: return a 1-letter name for vector j
 */
char lett( int j )
{
 if (j<26) return j+65;  // uppercase letters
 else
 if (j<52) return (j-26)+97; // lowercase letters
 else
 if (j<62) return (j-52)+48; // digits
 else
 //assert(0);  // no more printable letters for graphviz names
 // just to make it work around
 // (wrong visualization, but no crashed console output):
 return lett( j-62 );
}

// random integer between mi and ma (inclusive), equal distribution:
int rn( int mi, int ma )
{
 if (mi==ma) return mi;
 assert( mi < ma );
 int r = rand();
 //printf("r=%d\n", r );
 r = abs(r);
 r = r % (ma - mi + 1);
 assert (r>=0);
 assert (r<=ma-mi);
 return mi + r;
}


// special DAG nodes for convenience:
int vout = INIT_K;

int POTRF ( int id, int vin ) 
{ 
    assert( id >= 0 && id < N ); 
    assert( vin < MAX_K ); 
    name[id] = "POTRF"; 
    indeg[id] = outdeg[id] = 1;
    in[id][0] = vin;
    out[id][0] = vout++;
    return out[id][0];
}

int TRSM ( int id, int vin ) 
{ 
    assert( id >= 0 && id < N ); 
    assert( vin < MAX_K ); 
    name[id] = "TRSM"; 
    indeg[id] = outdeg[id] = 1;
    in[id][0] = vin;
    out[id][0] = vout++;
    return out[id][0];
}

int TRSM2 ( int id, int vin1, int vin2 ) 
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K && vin2 < MAX_K ); 
    name[id] = "TRSM"; 
    indeg[id] = 2; outdeg[id] = 1;
    in[id][0] = vin1;
    in[id][1] = vin2;
    out[id][0] = vout++;
    return out[id][0];
}

int SYRK2 ( int id, int vin1, int vin2 ) 
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K && vin2 < MAX_K ); 
    name[id] = "SYRK"; 
    indeg[id] = 2; outdeg[id] = 1;
    in[id][0] = vin1;
    in[id][1] = vin2;
    out[id][0] = vout++;
    return out[id][0];
}

int SYRK1 ( int id, int vin ) // clone for same input vector in both positions
{ 
    assert( id >= 0 && id < N ); 
    assert( vin < MAX_K ); 
    name[id] = "SYRK"; 
    indeg[id] = outdeg[id] = 1;
    in[id][0] = vin;
    out[id][0] = vout++;
    return out[id][0];
}

// GEMM: C_new = C_old + A x B
int GEMM ( int id, int vin1, int vin2, int vin3 ) 
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K && vin2 < MAX_K && vin3 < MAX_K ); 
    name[id] = "GEMM"; 
    indeg[id] = 3; outdeg[id] = 1;
    in[id][0] = vin1;
    in[id][1] = vin2;
    in[id][2] = vin3;
    out[id][0] = vout++;
    return out[id][0];
}

// GEMM2: C = A x B
int GEMM2 ( int id, int vin1, int vin2 ) 
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K && vin2 < MAX_K ); 
    name[id] = "GEMM"; 
    indeg[id] = 2; outdeg[id] = 1;
    in[id][0] = vin1;
    in[id][1] = vin2;
    out[id][0] = vout++;
    return out[id][0];
}

int MADD ( int id, int vin1, int vin2 ) 
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K && vin2 < MAX_K ); 
    name[id] = "MADD"; 
    indeg[id] = 2; outdeg[id] = 1;
    in[id][0] = vin1;
    in[id][1] = vin2;
    out[id][0] = vout++;
    return out[id][0];
}

int VSCALE ( int id, int vin1 ) 
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K ); 
    name[id] = "VSCALE"; 
    indeg[id] = 1; outdeg[id] = 1;
    in[id][0] = vin1;
    out[id][0] = vout++;
    return out[id][0];
}

int SAXPY ( int id, int vin1, int vin2 ) 
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K && vin2 < MAX_K ); 
    name[id] = "SAXPY"; 
    indeg[id] = 2; outdeg[id] = 1;
    in[id][0] = vin1;
    in[id][1] = vin2;
    out[id][0] = vout++;
    return out[id][0];
}

int SAXPY1( int id, int vin1 ) 
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K ); 
    name[id] = "SAXPY"; 
    indeg[id] = 1; outdeg[id] = 1;
    in[id][0] = vin1;
    out[id][0] = vout++;
    return out[id][0];
}

int SPMV1( int id, int vin1 ) // only 1 operand modeled, rest already on GPU
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K ); 
    //name[id] = "SPMV";  // buggy - for now replaced by the generic Kernel:
    name[id] = "Kernel"; 
    indeg[id] = 1; 
    outdeg[id] = 1;
    in[id][0] = vin1;
    out[id][0] = vout++;
    return out[id][0];
}

void SDOT ( int id, int vin1, int vin2 ) // no output vector
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K && vin2 < MAX_K ); 
    name[id] = "SDOT"; 
    indeg[id] = 2; outdeg[id] = 0;
    in[id][0] = vin1;
    in[id][1] = vin2;
}

void SDOT1 ( int id, int vin1 )    // square dot  - no output vector
{ 
    assert( id >= 0 && id < N ); 
    assert( vin1 < MAX_K ); 
    name[id] = "SDOT"; 
    indeg[id] = 1; outdeg[id] = 0;
    in[id][0] = vin1;
}

void calculate_live( void )  // later, parameter sched?
{
 int i, j;
 // clear live:
 for (i=0; i<N; i++) 
    for (j=0; j<INIT_K; j++) 
       live[i][j] = 0;

 // At entry of kernel 0, all vectors 0...INIT_K-1 are alive:
 for (j=0; j<INIT_K; j++)
    live[0][j] = 1;

 // now go through the schedule of kernels: 
 for (i=1; i<N; i++) { // later, use sched[i] below:

    // 1. those j alive at entry of i-1 are still alive at i
    //    if i-1 was not their last use
    //    (we do not yet consider downwards-exposed uses (roots) for simplicity)
    for (j=0; j<MAX_K; j++) {
       int j_is_alive = 0; 
       if (live[i-1][j]) {
          j_is_alive = 1;
          int ii, jj; // v------ N+1 to consider live-on-exit vectors
          for (ii=i; ii<N+1; ii++) { // check if j == in[ii][jj] for some (ii,jj)
             for (jj=0; jj<indeg[ii]; jj++) {
                if (j == in[ii][jj]) { 
                   goto FOUND;   // can stop searching 
                }
             }
          }
          j_is_alive = 0;  // if arriving here, not found
       }
       FOUND:
       if (j_is_alive)
          live[i][j] = 1;
    }
    // 2. out[i-1][0..outdeg[i-1]-1] are getting alive after i-1:
    for (j=0; j<outdeg[i-1]; j++)
       live[i][out[i-1][j]] = 1;
 }
 // 3. last row: add live-on-exit set to live[N][] for completeness:
 for (j=0; j<indeg[N]; j++)
    live[N][in[N][j]] = 1;

#if DEBUG
 // check output:
 printf("live matrix:");
 for (i=0; i<N+1; i++) { // later, use sched[i] below:
    printf("\n%3d:   ", i); 
    for (j=0; j<MAX_K; j++) 
       printf("%d", live[i][j]);
 }
 printf("\n");
#endif
}


/* samecolor: Trivial coloring of k vectors with k colors.
 * Output: kk, color[0...k-1] for each vector,
 *         and reverse mapping vec[][], nvec[] from colors to vectors
 */
void samecolor( void )
{
 int j;
 for (j=0; j<k; j++) {
    color[j] = j;
    nvec[j] = 1;
    vec[j][0] = j;
 }
 kk = k;
#if DEBUG
 printf("Colors assigned by samecolor, kk = %d\n", kk );
#endif
}

/* linearscan:  do a linear scan of the schedule 0...N-1.
 * Input: live matrix
 * Does: fold the k SSA vectors into kk <= k memory vectors (colors)
 * Output: kk, color[0...k-1] for each vector,
 *         and reverse mapping vec[][], nvec[] from colors to vectors
 */
void linearscan( void )
{
 int i, j, c;
 kk = INIT_K;
 int usedcolor[MAX_K+1]; // tmp bitvector, 1 iff color j currently in use  
 
 // initially, no colors assigned:
 for (c=0; c<MAX_K; c++) {
    nvec[c] = 0;
    usedcolor[c] = 0;
 }

 // fill in the input vectors:
 for (c=0; c<INIT_K; c++) {
    vec[c][nvec[c]++] = c;
    color[c] = c; 
    usedcolor[c] = 1; 
 } 
 // for temp vectors, find the first free color:
 for (i=0; i<N; i++) {
    //printf(" .. Kernel %d:\n", i );
    // last uses at i are freed out from usedcolor:
    for (j=0; j<indeg[i]; j++) {
       // in[i][j] is used in kernel i. Check if it stops living here:
       assert( live[i][in[i][j]] ); // sanity check
       if (!live[i+1][in[i][j]] ) { // dies here: free color
          assert( usedcolor[color[in[i][j]]] == 1); // sanity check
          //printf(" .... free color %d of vector %c\n",
          //         color[in[i][j]], lett(in[i][j]) ); 
          usedcolor[color[in[i][j]]] = 0;
       }
    }

    // new defs of i are included: 
    for (j=0; j<outdeg[i]; j++) {
       // out[i][j] starts living at i. Find a free color for it,
       // either reuse a color of a vector that is no longer live,
       // or take a fresh color (kk++).
       int jj; 
       for (jj=0; jj<MAX_K; jj++) {
          if (!usedcolor[jj]) { // first free color found:
             color[out[i][j]] = jj;
             usedcolor[jj] = 1;
             //printf("  .. Assigned color %d to %c\n", jj, lett(out[i][j])); 
             if (jj>kk) kk=jj; // it was a fresh color
             goto FOUND3;
          } 
       }
       assert(0); // should always find some color
       FOUND3:
         ;
    }
 }
#if DEBUG
 printf("Colors assigned by linearscan, kk = %d\n", kk );
#endif
}


void writedot_DAG( char *fname )
{
 int i, j;
 FILE *fp;
 if ((fp = fopen(fname,"w"))==NULL)
    fprintf(stderr, "Error: cannot open file %s\n", fname);
#if DEBUG
 printf("writing dot file named %s\n", fname ); 
#endif
 // emit preamble:
 fprintf(fp, "digraph {\n");

 for (i=0; i<N; i++) { // Kernel i
    if (device[i]==0) // CPU: 
       fprintf(fp, " %s_%d [fontsize=24 style=filled fillcolor=yellow]\n", name[i], i);
    else
       fprintf(fp, " %s_%d [fontsize=24 style=filled fillcolor=deepskyblue]\n", name[i], i);
 }
 for (j=0; j<INIT_K; j++) { // Input vector j
   assert (location[j]>=1);
   if (location[j]==1)
     fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=yellow shape=box]\n", lett(j));
     // fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=aliceblue shape=box]\n", lett(j));
   else if (location[j]==2)
     fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=cyan shape=box]\n", lett(j));
     // fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=darkseagreen3 shape=box]\n", lett(j));
   else
     fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=greenyellow shape=box]\n", lett(j));
     // fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=turquoise shape=box]\n", lett(j));
    
 }
 for (j=INIT_K; j<k; j++) // Vector j
 { char *bordercolor="blue";
   int l;
   for (l=0; l<indeg[N]; l++)
      if( in[N][l]==j ) bordercolor="red"; // live on exit
   if (location[j]==1)
     fprintf(fp, " %c [color=%s fontsize=24 style=filled fillcolor=yellow shape=box]\n", lett(j), bordercolor );
     // fprintf(fp, " %c [color=%s fontsize=24 style=filled fillcolor=aliceblue shape=box]\n", lett(j));
   else if (location[j]==2)
     fprintf(fp, " %c [color=%s fontsize=24 style=filled fillcolor=cyan shape=box]\n", lett(j), bordercolor);
     // fprintf(fp, " %c [color=blue fontsize=24 style=filled fillcolor=darkseagreen3 shape=box]\n", lett(j));
   else
     fprintf(fp, " %c [color=%s fontsize=24 style=filled fillcolor=greenyellow shape=box]\n", lett(j), bordercolor);
     // fprintf(fp, " %c [color=blue fontsize=24 style=filled fillcolor=turquoise shape=box]\n", lett(j));
   }
 
 // emit edges:
 for (i=0; i<N; i++) { // Kernel i
   for (j=0; j<indeg[i]; j++) {
     int v=in[i][j];
     int vloc = location[v]; // 1=CPU,2=GPU,3=both, 0=not allocated
     assert( vloc > 0); 
     if (vloc>=2 && device[i]==0)  // transfer download:
       fprintf(fp, "  %c -> %s_%d [penwidth=2 color=black]\n", lett(in[i][j]), name[i], i);
     else if (vloc==1 && device[i]==1) // transfer upload: 
       fprintf(fp, "  %c -> %s_%d [penwidth=2 color=black]\n", lett(in[i][j]), name[i], i);
     else  // no transfer along this edge:
       fprintf(fp, "  %c -> %s_%d [penwidth=2 color=black]\n", lett(in[i][j]), name[i], i);
   }
   for (j=0; j<outdeg[i]; j++) {
     int v=out[i][j];
     int vloc = location[v]; // 1=CPU,2=GPU,3=both, 0=not allocated
     assert( vloc > 0); 
     assert( ncons[v]>=1 );
     if (device[i]==0)  // no transfer yet, data produced into main memory:
        fprintf(fp, "  %s_%d -> %c [penwidth=2 color=black]\n", name[i], i, lett(out[i][j]));
     else // i on GPU:
       if (vloc==3) // produced data remains on GPU, no transfer
         fprintf(fp, "  %s_%d -> %c [penwidth=2 color=black]\n", name[i], i, lett(out[i][j]));
       else  // both keep and download along this edge:
         fprintf(fp, "  %s_%d -> %c [penwidth=2 color=black]\n", name[i], i, lett(out[i][j]));
   }
 } 
 // emit suffix:
 fprintf(fp, "}\n");
 fclose(fp);
}

    
void writedot_Aff( char *fname )
{
 int j, jj;
 FILE *fp;
 if ((fp = fopen(fname,"w"))==NULL)
    fprintf(stderr, "Error: cannot open file %s\n", fname);
#if DEBUG
 printf("writing dot file named %s\n", fname ); 
#endif
 // emit preamble:
 fprintf(fp, "strict graph {\n");

 for (j=0; j<INIT_K; j++) { // Input vector j
   assert (location[j]>=1);
   if (location[j]==1)
     fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=yellow shape=box]\n", lett(j));
     // fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=aliceblue shape=box]\n", lett(j));
   else if (location[j]==2)
     fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=cyan shape=box]\n", lett(j));
     // fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=darkseagreen3 shape=box]\n", lett(j));
   else
     fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=greenyellow shape=box]\n", lett(j));
     // fprintf(fp, " %c [color=black fontsize=24 style=filled fillcolor=turquoise shape=box]\n", lett(j));
 }
 for (j=INIT_K; j<k; j++) // Vector j
   if (location[j]==1)
     fprintf(fp, " %c [color=blue fontsize=24 style=filled fillcolor=yellow shape=box]\n", lett(j));
     // fprintf(fp, " %c [color=blue fontsize=24 style=filled fillcolor=aliceblue shape=box]\n", lett(j));
   else if (location[j]==2)
     fprintf(fp, " %c [color=blue fontsize=24 style=filled fillcolor=cyan shape=box]\n", lett(j));
     // fprintf(fp, " %c [color=blue fontsize=24 style=filled fillcolor=darkseagreen3 shape=box]\n", lett(j));
   else
     fprintf(fp, " %c [color=blue fontsize=24 style=filled fillcolor=greenyellow shape=box]\n", lett(j));
     // fprintf(fp, " %c [color=blue fontsize=24 style=filled fillcolor=turquoise shape=box]\n", lett(j));
 
 // emit edges:
 for (j=0; j<k; j++) {
    for (jj=0; jj<k; jj++) {
       if (j==jj) continue;
       if (aff[j][jj]) // affinity edge to draw
         if (hamilton[bestweightedhamilton][j][jj]>0 || hamilton[bestweightedhamilton][jj][j]>0)
           fprintf(fp, " %c -- %c [penwidth=%d color=red]\n", lett(j), lett(jj), (int)(2*(-1+2*aff[j][jj])) );  // edge (j,jj) 
         else
           fprintf(fp, " %c -- %c [penwidth=%d]\n", lett(j), lett(jj), (int)(2*(-1+2*aff[j][jj])) );  // edge (j,jj) 
    }
 }
 // emit suffix:
 fprintf(fp, "}\n");
 fclose(fp);
}


// LIVEONEXIT marks vectors live on exit.
#define LIVEONEXIT( j ) { \
	in[N][indeg[N]++] = j; \
}

/* build a DAG with parameters N, MAX_K, MAX_INDEG, MAX_OUTDEG:
 */
void build_DAG( void ) 
{
 int i, j, t, done=0, src;
 assert( INIT_K >= MAX_INDEG );
 printf("Generating a trace program (N=%d, INIT_K=%d ...\n", N, INIT_K);
 indeg[N] = 0; // add with LIVEONEXIT()
 device[N] = 0; // CPU as LIVEONEXIT should be in main memory
   
#ifndef SPECIALDAG
 // Synthetic DAG. Set name and device for each kernel:
 for (i=0; i<N; i++) {
    name[i] = "Kernel";  // generic synthetic kernel
    device[i] = rn( 0, 1 ); // 0=CPU, 1=GPU
 }
 name[N] = "LiveOnExit";  // artificial kernel reading live-on-exit vectors
 device[N] = 0; // vectors live on exit must be in main memory

 // set call parameter degrees:
 for (i=0; i<N; i++) {
    indeg[i] = rn( MIN_INDEG, MAX_INDEG );
    for (j=0; j<indeg[i]; j++) {
       // choose src, avoid duplicates:
       do {
          // favor recently computed vectors as inputs:
          int lb_scope = 0; // max( 0, k-2*MAX_INDEG-MAX_OUTDEG );
          src = rn( lb_scope, k-1 );
          done = 1;
          for (t=0; t<j; t++)
            if (src == in[i][t])  done=0;
       } while (!done);
       in[i][j] = src; 
       //printf("in[%d][%d] = %d\n", i, j, in[i][j]);  
    } 
    outdeg[i] = rn( MIN_OUTDEG, MAX_OUTDEG );
    // single-assignment: fresh new vector produced by each kernel call:
    for (j=0; j<outdeg[i]; j++) {
       out[i][j] = k++; 
       //printf("out[%d][%d] = %d\n", i, j, out[i][j]);  
    }
    //printf("degrees(%d)=(%d,%d)\n", i, indeg[i], outdeg[i] );
 }
#else

#if SPECIALDAG==1    
 // Horner-DAG: linear chain of alternating accumulating VMUL and VADD,
 // where all VMULs take the same second operand.
 assert(MAX_INDEG >= 2);
 assert(INIT_K == N+2);
 assert(MAX_K >= 2*N+2);
 // inputs: v0=x_vals, v1...N = coeffs, accum. res = N+1,...3N
 for (i=0; i<N/2; i++) {  // N kernels, sequence VMUL, VADD, VMUL, VADD ...
    // VMUL
    name[2*i] = "VMUL";
    indeg[2*i] = 2;
    in[2*i][0] = N+1+2*i; // res (accumulator)
    in[2*i][1] = 0; // x_vals
    outdeg[2*i] = 1;
    out[2*i][0] = N+1+2*i+1; // res, next version
    // VADD
    name[2*i+1] = "VADD";
    indeg[2*i+1] = 2;
    in[2*i+1][0] = N+1+2*i+1; // res (accumulator)
    in[2*i+1][1] = 2*i+1; // coeffs[1...N] = vec 1...N
    outdeg[2*i+1] = 1;
    out[2*i+1][0] = N+1+2*i+2; // res, next version
 }
 k = 2*N+2; // #vectors: N+2 different inputs, N generated outputs
 assert( k<=MAX_K );
 // set device for each kernel:
 for (i=0; i<N; i++)
    device[i] = 1; // 0=CPU, 1=GPU
 // device[N] = 0; // vectors live on exit must be in main memory
 LIVEONEXIT( 2*N+1 );

#elif SPECIALDAG==2
 // Linear chain:
 assert(INIT_K == 1);
 assert(MAX_K >= N+1);
 // input: v0=image.  N=5 for face detection, N=6 for line detection...
 for (i=0; i<N; i++) {  // N kernels ...
     name[i] = "Kernel"; // generic synthetic kernel
     indeg[i]=outdeg[i]=1;
     in[i][0] = i;
     out[i][0] = i+1;
 }
 k = N+1; // #vectors: 1 input image, N generated outputs
 assert( k<=MAX_K );
 // set device for each kernel:
 for (i=0; i<N; i++)
    device[i] = 1; // all on GPU.  0=CPU, 1=GPU
 LIVEONEXIT( N );
 // device[N] = 0; // vectors live on exit must be in main memory

#elif SPECIALDAG==3
 // Linear chain with Random device assignment:
 assert(INIT_K == 1);
 assert(MAX_K >= N+1);
 // input: v0=image.  N=5 for face detection, N=6 for line detection...
 for (i=0; i<N; i++) {  // N kernels, sequence VMUL, VADD, VMUL, VADD ...
     name[i] = "Kernel"; // generic synthetic kernel
     indeg[i]=outdeg[i]=1;
     in[i][0] = i;
     out[i][0] = i+1;
 }
 k = N+1; // #vectors: 1 input image, N generated outputs
 assert( k<=MAX_K );
 // set device for each kernel:
 for (i=0; i<N; i++)
    device[i] = rn(0,1); // 50% GPU.  0=CPU, 1=GPU
 LIVEONEXIT( N );
 // device[N] = 0; // vectors live on exit must be in main memory

#elif SPECIALDAG==4
 // Outbound Tree, single input vector (v0), two outputs per kernel call
 assert(INIT_K == 1);
 assert(MAX_K >= 2*N+1);
 // N is a power of 2 minus 1, e.g. 15 (0..14): 
 for (i=0; i<N; i++) {
    name[i] = "Kernel"; // generic synthetic kernel
    indeg[i]=1; outdeg[i] = 2; 
    in[i][0] = i;
    out[i][0] = 2*i+1;
    out[i][1] = 2*i+2;
 }
 k = 2*N+1;
 for (i=0; i<N; i++)
    device[i] = 1; // rn(0,1); // 50% GPU.  0=CPU, 1=GPU
 // device[N] = 0; // vectors live on exit must be in main memory
 for (i=N; i<2*N+1; i++)
    LIVEONEXIT( i );  // vectors N+1,...,2N are live on exit

#elif SPECIALDAG==5
 // Outbound Tree Random-device, single input vector (v0), two outputs per kernel call
 assert(INIT_K == 1);
 assert(MAX_K >= 2*N+1);
 // N is a power of 2 minus 1, e.g. 15 (0..14): 
 for (i=0; i<N; i++) {
    name[i] = "Kernel"; // generic synthetic kernel
    indeg[i]=1; outdeg[i] = 2; 
    in[i][0] = i;
    out[i][0] = 2*i+1;
    out[i][1] = 2*i+2;
 }
 k = 2*N+1;
 for (i=0; i<N; i++)
    device[i] = rn(0,1); // 50% GPU.  0=CPU, 1=GPU
 // device[N] = 0; // vectors live on exit must be in main memory
 for (i=N; i<2*N+1; i++)
    LIVEONEXIT( i );  // vectors N+1,...,2N are live on exit

#elif SPECIALDAG==6
 // Inbound Tree, N+1 input vectors, single output vector (v0),
 //   two inputs and one output per kernel call
 assert(INIT_K == N+1);
 assert(MAX_K >= 2*N-1);
 // N is a power of 2 minus 1, e.g. 15 (0..14): 
 for (i=0; i<N; i++) {
    name[i] = "Kernel"; // generic synthetic kernel
    indeg[i]=2; outdeg[i] = 1; 
    in[i][0] = 2*i;
    in[i][1] = 2*i+1;
    out[i][0] = N+1+i;
 }
 k = 2*N+1;
 for (i=0; i<N; i++)
    device[i] = 1; // rn(0,1); // 50% GPU.  0=CPU, 1=GPU
 device[N] = 0; // vectors live on exit must be in main memory
 LIVEONEXIT( 2*N );  // vectors N,...,2N-1 are live on exit

#elif SPECIALDAG==7
 // Inbound Tree, Random device, N+1 input vectors, single output vector (v0),
 //   two inputs and one output per kernel call
 assert(INIT_K == N+1);
 assert(MAX_K >= 2*N-1);
 // N is a power of 2 minus 1, e.g. 15 (0..14): 
 for (i=0; i<N; i++) {
    name[i] = "Kernel"; // generic synthetic kernel
    indeg[i]=2; outdeg[i] = 1; 
    in[i][0] = 2*i;
    in[i][1] = 2*i+1;
    out[i][0] = N+1+i;
 }
 k = 2*N+1;
 for (i=0; i<N; i++)
    device[i] = rn(0,1); // 50% GPU.  0=CPU, 1=GPU
 // device[N] = 0; // vectors live on exit must be in main memory
 LIVEONEXIT( 2*N );

#elif SPECIALDAG==8
 // 4x4 blocks Left-looking Cholesky factorization, N==20
 assert(N == 20);
 assert(INIT_K == 10);
 assert(MAX_K >= 30);
 // global vout is set to INIT_K
 int a00=0, a10=1, a20=2, a30=3, a11=4, a21=5, a22=6, a31=7, a32=8, a33=9;
 int A00 = POTRF( 0, a00 );  // A00 = POTRF(CHOL)(A00) 
 int A10 = TRSM2 ( 1, a10, A00 );  // A10 = TRSM(A10,A00) 
 int A20 = TRSM2 ( 2, a20, A00 );  // A20 = TRSM(A20,A00) 
 int A30 = TRSM2 ( 3, a30, A00 );  // A30 = TRSM(A30,A00) 
 int A11 = SYRK2 ( 4, a11, A10 );     // A11 = SYRK(A10,A10) 
 int A22 = SYRK2 ( 5, a22, A20 );     
 int A21 = GEMM ( 6, a21, A20, A10 ); 
 int A33 = SYRK2 ( 7, a33, A30 );     
 int A31 = GEMM ( 8, a31, A30, A10 ); 
 int A32 = GEMM ( 9, a32, A30, A20 ); 
 int A11a = POTRF( 10, A11 );
 int A21a = TRSM2 ( 11, A21, A11a );
 int A31a = TRSM2 ( 12, A31, A11a );
 int A22a = SYRK2( 13, A22, A21a );     
 int A33a = SYRK2( 14, A33, A31a ); 
 int A32a = GEMM ( 15, A32, A31a, A21a ); 
 int A22b = POTRF( 16, A22a );
 int A32b = TRSM2( 17, A32a, A22b );
 int A33b = SYRK2( 18, A33a, A32b );
 int A33c = POTRF( 19, A33b );
#if 0
 // global vout is set to INIT_K
 int chol0 = POTRF( 0, 0 );  // A00 = POTRF(CHOL)(A00) 
 int trsm1 = TRSM2 ( 1, 1, chol0 );  // A10 = TRSM(A10,A00) 
 int trsm2 = TRSM2 ( 2, 2, chol0 );  // A20 = TRSM(A20,A00) 
 int trsm3 = TRSM2 ( 3, 3, chol0 );  // A30 = TRSM(A30,A00) 
 int syrk4 = SYRK1 ( 4, trsm1 );     // A11 = SYRK(A10,A10) 
 int gemm5 = GEMM ( 5, trsm1, trsm2 ); 
 int gemm6 = GEMM ( 6, trsm1, trsm3 ); 
 int syrk7 = SYRK1 ( 7, trsm2 );     
 int gemm8 = GEMM ( 8, trsm2, trsm3 ); 
 int syrk9 = SYRK1 ( 9, trsm3 );     
 int chol10= POTRF( 10, syrk4 );
 int trsm11= TRSM2 ( 11, chol10, gemm5 );
 int trsm12= TRSM2 ( 12, chol10, gemm6 );
 int syrk13 = SYRK2( 13, trsm11, syrk7 );     
 int gemm14 = GEMM ( 14, trsm11, trsm12 ); 
 int syrk15 = SYRK2( 15, trsm12, syrk9 ); 
 int chol16 = POTRF( 16, syrk13 );
 int trsm17 = TRSM2( 17, chol16, gemm14 );
 int syrk18 = SYRK2( 18, trsm17, syrk15 );
 int chol19 = POTRF( 19, syrk18 );
#endif
 k = 30; // N+INIT_K;
 // TRSM, SYRK and GEMM run on GPU using CUBLAS (in PLASMA, Sect. 5.1 in Ltaief et al.#223) 
 for (i=0; i<N; i++)
    device[i] = 1; // GPU
 device[A00-INIT_K] = device[A11a-INIT_K] = device[A22b-INIT_K] = device[A33c-INIT_K] = 0; // CPU 
 LIVEONEXIT( A00 );
 LIVEONEXIT( A10 );
 LIVEONEXIT( A20 );
 LIVEONEXIT( A30 );
 LIVEONEXIT( A11a );
 LIVEONEXIT( A21a );
 LIVEONEXIT( A31a );
 LIVEONEXIT( A22b );
 LIVEONEXIT( A32b );
 LIVEONEXIT( A33c );
 // device[N] = 0; // vectors live on exit must be in main memory

#elif SPECIALDAG==9
 // 4x4 blocks Left-looking Cholesky factorization, N==20, force SYRK on CPU
 assert(N == 20);
 assert(INIT_K == 10);
 assert(MAX_K >= 24);
 // global vout is set to INIT_K
 int a00=0, a10=1, a20=2, a30=3, a11=4, a21=5, a22=6, a31=7, a32=8, a33=9;
 int A00 = POTRF( 0, a00 );  // A00 = POTRF(CHOL)(A00) 
 int A10 = TRSM2 ( 1, a10, A00 );  // A10 = TRSM(A10,A00) 
 int A20 = TRSM2 ( 2, a20, A00 );  // A20 = TRSM(A20,A00) 
 int A30 = TRSM2 ( 3, a30, A00 );  // A30 = TRSM(A30,A00) 
 int A11 = SYRK2 ( 4, a11, A10 );     // A11 = SYRK(A10,A10) 
 int A22 = SYRK2 ( 5, a22, A20 );     
 int A21 = GEMM ( 6, a21, A20, A10 ); 
 int A33 = SYRK2 ( 7, a33, A30 );     
 int A31 = GEMM ( 8, a31, A30, A10 ); 
 int A32 = GEMM ( 9, a32, A30, A20 ); 
 int A11a = POTRF( 10, A11 );
 int A21a = TRSM2 ( 11, A21, A11a );
 int A31a = TRSM2 ( 12, A31, A11a );
 int A22a = SYRK2( 13, A22, A21a );     
 int A33a = SYRK2( 14, A33, A31a ); 
 int A32a = GEMM ( 15, A32, A31a, A21a ); 
 int A22b = POTRF( 16, A22a );
 int A32b = TRSM2( 17, A32a, A22b );
 int A33b = SYRK2( 18, A33a, A32b );
 int A33c = POTRF( 19, A33b );
#if 0
 // global vout is set to INIT_K
 int chol0 = POTRF( 0, 0 );  // A00 = POTRF(CHOL)(A00) 
 int trsm1 = TRSM2 ( 1, 1, chol0 );  // A10 = TRSM(A10,A00) 
 int trsm2 = TRSM2 ( 2, 2, chol0 );  // A20 = TRSM(A20,A00) 
 int trsm3 = TRSM2 ( 3, 3, chol0 );  // A30 = TRSM(A30,A00) 
 int syrk4 = SYRK1 ( 4, trsm1 );     // A11 = SYRK(A10,A10) 
 int gemm5 = GEMM ( 5, trsm1, trsm2 ); 
 int gemm6 = GEMM ( 6, trsm1, trsm3 ); 
 int syrk7 = SYRK1 ( 7, trsm2 );     
 int gemm8 = GEMM ( 8, trsm2, trsm3 ); 
 int syrk9 = SYRK1 ( 9, trsm3 );     
 int chol10= POTRF( 10, syrk4 );
 int trsm11= TRSM2 ( 11, chol10, gemm5 );
 int trsm12= TRSM2 ( 12, chol10, gemm6 );
 int syrk13 = SYRK2( 13, trsm11, syrk7 );     
 int gemm14 = GEMM ( 14, trsm11, trsm12 ); 
 int syrk15 = SYRK2( 15, trsm12, syrk9 ); 
 int chol16 = POTRF( 16, syrk13 );
 int trsm17 = TRSM2 ( 17, chol16, gemm14 );
 int syrk18 = SYRK2( 18, trsm17, syrk15 );
 int chol19 = POTRF( 19, syrk18 );
#endif
 k = 30; // N+INIT_K;
 // TRSM and GEMM run on GPU using CUBLAS (in PLASMA, Sect. 5.1 in Ltaief et al.#223) 
 for (i=0; i<N; i++)
    device[i] = 1; // GPU
 device[A00-INIT_K] = device[A11a-INIT_K] = device[A22b-INIT_K] = device[A33c-INIT_K] = 0; // CPU 
 // As variation, here also SYRK2 calls on GPU:
 device[A11-INIT_K] = device[A22-INIT_K] = device[A33-INIT_K] = device[A22a-INIT_K] = device[A33a-INIT_K] = device[A33b-INIT_K] = 0; // CPU 
 LIVEONEXIT( A00 );
 LIVEONEXIT( A10 );
 LIVEONEXIT( A20 );
 LIVEONEXIT( A30 );
 LIVEONEXIT( A11a );
 LIVEONEXIT( A21a );
 LIVEONEXIT( A31a );
 LIVEONEXIT( A22b );
 LIVEONEXIT( A32b );
 LIVEONEXIT( A33c );

#elif SPECIALDAG==10
// 2x2 blocks Matrix-matrix multiply 
 assert(N == 12); // 8 MatMuls and 4 Madds
 assert(INIT_K == 8); // 2x2 blocks of A (0..3) and B (4..7)
 // global vout is set to INIT_K
 assert(MAX_K >= 20); // 8+12
 k = 20;
 int A00 = 0;
 int A01 = 1;
 int A10 = 2;
 int A11 = 3;
 int B00 = 4;
 int B01 = 5;
 int B10 = 6;
 int B11 = 7;
 int MM0 = GEMM2( 0, A00, B00 );
 int MM1 = GEMM2( 1, A01, B10 );
 int C00 = MADD( 2, MM0, MM1 );
 int MM2 = GEMM2( 3, A00, B01 );
 int MM3 = GEMM2( 4, A01, B11 );
 int C01 = MADD( 5, MM2, MM3 );
 int MM4 = GEMM2( 6, A10, B00 );
 int MM5 = GEMM2( 7, A11, B10 );
 int C10 = MADD( 8, MM4, MM5 );
 int MM6 = GEMM2( 9, A10, B01 );
 int MM7 = GEMM2(10, A11, B11 );
 int C11 = MADD(11, MM6, MM7 );
 // let the GEMM run on GPU, MADD on CPU:
 for (i=0; i<N; i++)
    device[i] = 1; // GPU
 device[2] = device[5] = device[8] = device[11] = 0; // CPU
 LIVEONEXIT(C00);
 LIVEONEXIT(C01);
 LIVEONEXIT(C10);
 LIVEONEXIT(C11);

#elif SPECIALDAG==11
 // Conjugate Gradient Loop steady state DFG
 //  see also CUDA samples in CUDA 10 (6_Advanced), but does not work with CUDA 8
 assert(N == 7); // kernels
 assert(INIT_K == 2); // r and p, the other vectors are already on GPU
 // global vout is set to INIT_K
 assert(MAX_K >= 7); // 2+5 (the 2 SDOT kernels do not produce any vector)
 k = 7;
 int v_r = 0;
 int v_p = 1;
 int v_p1 = VSCALE( 0, v_p );
 int v_p2 = SAXPY( 1, v_r, v_p1 );
 int v_Ax = SPMV1( 2, v_p2 );
            SDOT( 3, v_Ax, v_p2 );
 int v_r1 = SAXPY( 4, v_Ax, v_r );
 int v_x  = SAXPY1( 5, v_p2 );
            SDOT1( 6, v_r1 );
 for (i=0; i<N; i++)
    device[i] = 1; // GPU
 device[2] = 0; // SPMV1 on CPU - arbitrary choice
 LIVEONEXIT( v_r1 );
 LIVEONEXIT( v_p2 );
 LIVEONEXIT( v_x );

#elif SPECIALDAG==12
 // Conjugate Gradient Loop steady state DFG
 //  see also CUDA samples in CUDA 10 (6_Advanced), but does not work with CUDA 8
 assert(N == 7); // kernels
 assert(INIT_K == 2); // r and p, the other vectors are already on GPU
 // global vout is set to INIT_K
 assert(MAX_K >= 7); // 2+5 (the 2 SDOT kernels do not produce any vector)
 k = 7;
 int v_r = 0;
 int v_p = 1;
 int v_p1 = VSCALE( 0, v_p );
 int v_p2 = SAXPY( 1, v_r, v_p1 );
 int v_Ax = SPMV1( 2, v_p2 );
            SDOT( 3, v_Ax, v_p2 );
 int v_r1 = SAXPY( 4, v_Ax, v_r );
 int v_x  = SAXPY1( 5, v_p2 );
            SDOT1( 6, v_r1 );
 for (i=0; i<N; i++)
    device[i] = 1; // GPU
 device[2] = 0; // SPMV1 on CPU - arbitrary choice
 device[3] = 0; // SDOT1 on CPU - arbitrary choice
 device[6] = 0; // SDOT1 on CPU - arbitrary choice
 LIVEONEXIT( v_r1 );
 LIVEONEXIT( v_p2 );
 LIVEONEXIT( v_x );

#endif
#endif

#ifndef SPECIALDAG
 // Synthetic DAG: determine vectors live on exit:
 // set in[N] and indeg[N] to those vectors that occur in out
 // of some i=0...N-1 but not in an in[] in subsequent kernels.
 // indeg[N] = 0;
 for (i=0; i<N; i++) {
    for (j=0; j<outdeg[i]; j++) {
       int v = out[i][j]; // candidate
       int v_is_live_on_exit = 0;
       int ii, jj; 

       // first check if v had not already been added earlier:
       for (jj=0; jj<indeg[N]; jj++)
          if (v == in[N][jj])
             goto FOUND2;  // skip adding it again
          
       // now check if v occurs in any subsequent in[] set:
       v_is_live_on_exit = 1;
       for (ii=i; ii<N; ii++) {
          for (jj=0; jj<indeg[ii]; jj++) {
             if (v == in[ii][jj]) { // found:
                v_is_live_on_exit = 0;
                // printf(" ---- found v=%d non live on exit\n", v);
                goto FOUND2;
             }
          } 
       }
       // if arriving here, not found --> add it
       FOUND2:
       if (v_is_live_on_exit) 
          in[N][indeg[N]++] = v;  // add to live on exit set
    }
 }
#endif
 assert( indeg[N] > 0 );  // sanity check

 // now build the dual graph:
 for (j=0; j<MAX_K; j++) // init nprod, ncons:
   nprod[j] = ncons[j] = 0;

 for (i=0; i<N+1; i++) {
    for (j=0; j<indeg[i]; j++) {
       int v = in[i][j];
       cons[v][ncons[v]++] = i;
    }
 }
 for (i=0; i<N; i++) {
    for (j=0; j<outdeg[i]; j++) {
       int v = out[i][j];
       prod[v][nprod[v]++] = i;
    }
 }

 // finally, set the vector locations (1=CPU, 2=GPU, 3=both, 0=undef):
 for (j=0; j<INIT_K; j++) {  // input vectors in CPU memory
    int jj;
    location[j]=1;  // input vectors already in CPU memory
    for (jj=0; jj<ncons[j]; jj++)
       if (device[cons[j][jj]]==1) // any consumer kernel on GPU?
          location[j]=3;  // also copied to GPU memory
 }

 for (j=INIT_K; j<k; j++) { // vector j:
    int jj;
    location[j] = 0; // default 
    // assert( nprod[j] == 1); // sanity check // 190224 TESTWEISE RAUS!!!
    for (jj=0; jj<nprod[j]; jj++) {
       i = prod[j][jj];
       if (device[i]==0) // CPU kernel
          location[j] |=1;  // allocate on CPU memory
       else
          location[j] |=2;  // allocate on GPU memory
    }
    for (jj=0; jj<ncons[j]; jj++) {
       i = cons[j][jj];
       if (device[i]==0) // CPU kernel
          location[j] |=1;  // allocate on CPU memory
       else
          location[j] |=2;  // allocate on GPU memory
    }
    //printf(" Location[vector %c]=%d\n", lett(j), location[j] );
 }
}


/* calculate_aff:  
 * Input: vector coloring by kk, color[], vec[][], nvec[].
 * Output: kk x kk affinity matrix aff[][]
 *         with aff[i][j] indicating the benefit of mapping
 *         colors i and j to consecutive places in memory.
 */
void calculate_aff( void )
{
 int i, j, jj;
 // init aff array:
 for (i=0; i<MAX_K; i++)
    for (j=0; j<MAX_K; j++)
       aff[i][j] = 0.0; 

 // plane sweep through the vectors' _earliest and _latest ranges:
  for (j=0; j<k; j++) {
     for (jj=0; jj<k; jj++) { // check for affinity (j,jj) in step i:
        if (j==jj) continue; 
        if (up_earliest[j]>=0 && up_earliest[jj]>=0) { // both must to GPU:
           // check for overlap of up_ intervals:
           if (overlap( up_earliest[j], up_latest[j], 
                        up_earliest[jj], up_latest[jj])) {
              aff[j][jj] ++;  
              if (up_latest[j]==up_latest[jj]) // siblings
                 aff[j][jj] += EPSILON_UP;  
           }
        }
        if (down_latest[j]<N+2 && down_latest[jj]<N+2) {
           // check for overlap of down_ intervals:
           if (overlap( down_earliest[j], down_latest[j], 
                        down_earliest[jj], down_latest[jj])) {
              aff[j][jj] ++;  
              // if (down_earliest[j]==down_earliest[jj]) // siblings
              if (down_earliest[j]==down_earliest[jj]) // siblings
                 aff[j][jj] += EPSILON_DOWN;  
           }
        }
    }
 }

 // add preset affinities for certain data structures in examples:
#if SPECIALDAG==10
 for (j=0; j<3; j++)   // Assumption: A00...A11 consecutive in memory
    aff[j][j+1] = aff[j+1][j] = 8; // LARGE
    //for (jj=0; jj<4; jj++)
    //   aff[j][jj] = 8; // LARGE
 for (j=4; j<8; j++)   // B00...B11
    for (jj=4; jj<8; jj++)
       aff[j][jj] = 8; // LARGE
#endif
 
#if DEBUG
 // check output:
 printf("Affinity matrix aff[][]:");
 for (i=0; i<kk; i++) {
    printf("\n%c:   ", lett(i)); 
    for (j=0; j<kk; j++) 
       printf(" %1.1f", aff[i][j]);
 }
 printf("\n");
#endif
}

/* calculate_aff_old:  
 * Input: vector coloring by kk, color[], vec[][], nvec[].
 * Output: kk x kk affinity matrix aff[][]
 *         with aff[i][j] indicating the benefit of mapping
 *         colors i and j to consecutive places in memory.
 */
void calculate_aff_old( void )
{
 int i, j, ii, jj, jjj;
 // init aff array:
 for (i=0; i<MAX_K; i++)
    for (j=0; j<MAX_K; j++)
       aff[i][j] = 0; 

 // go through the schedule:
 for (i=0; i<N; i++) {
    if (indeg[i]>1) {
       for (j=0; j<indeg[i]; j++) {
          if (in[i][j] >= INIT_K) continue; // skip intermediate vectors
          for (jj=0; jj<indeg[i]; jj++) {
             if (in[i][jj] >= INIT_K) continue; // skip intermediate vectors
             if (jj!=j) {
                // (in.j,in.jj) benefits if both are uploaded 
                // simultaneously, i.e., are first-uses here:
                // i.e. no use of inij, inijj before i:
                int both_first_uses = 1;
                for (ii=0; ii<i; ii++) {
                   for (jjj=0; jjj<indeg[ii]; jjj++) { // earlier uses
                      if ( in[ii][jjj] == in[i][j]
                          || in[ii][jjj] == in[i][jj] ) {
                         both_first_uses = 0;
                         goto FOUND4;
                      }
                   }
                }
                FOUND4: 
                if (both_first_uses)
                   aff[in[i][j]][in[i][jj]] ++;
             }
          }
       }
    }
 }
#if DEBUG
 // check output:
 printf("Affinity matrix aff[][]:");
 for (i=0; i<kk; i++) {
    printf("\n%c:   ", lett(i)); 
    for (j=0; j<kk; j++) 
       printf("%2.1f", aff[i][j]);
 }
 printf("\n");
#endif
}


/* earliestlatest(): 
 * calculate up_earliest, up_latest, down_earliest, down_latest
 * given DAG and schedule.
 * Needed for calculating affinities.
 */
void earliestlatest() 
{
 int i, j, jj;
 // set some default values:
 for (j=0; j<MAX_K; j++) {
    up_earliest[j] = down_earliest[j] = -1; // magic, below lower bound
    up_latest[j] = down_latest[j] = N+2; // magic, above upper bound
 }

 // init variables can be uploaded immediately, if needed:
 for (j=0; j<INIT_K; j++) {
    // up_earliest[j] = 0; // available before kernel 0  // 181213 moved below:
    // find all GPU consumers of vector j:
    for (jj=0; jj<ncons[j]; jj++) {
        if (device[cons[j][jj]]==1) { // GPU consumer found:
           up_earliest[j] = 0; // available before kernel 0
           up_latest[j] = min( up_latest[j], cons[j][jj] );
        }
    }
    // no point in setting down_..., as 0..INIT_K-1 all initially reside in main memory.
 }
 for (j=INIT_K; j<k; j++) { // vectors computed by kernels:
    // find the producer of j:
    // assert( nprod[j] == 1);  // sanity check  // 190224 TESTWEISE RAUS!!!
    if ( nprod[j] ) {
       int pj = prod[j][0]; // the producer of j
       if (device[pj]==1) { // produced on GPU:
          int ncpuconsumers = 0;
          int earliestcpuuse = N+2; 
          for (jj=0; jj<ncons[j]; jj++) { // find CPU consumers:
             if (device[cons[j][jj]] == 0) {
                ncpuconsumers++; 
                //printf("  TEST cons[%c][%d] = %d\n", lett(j), jj, cons[j][jj] );
                earliestcpuuse = min( earliestcpuuse, cons[j][jj] );
             }
          }
          if (ncpuconsumers > 0) { // download necessary
             assert(earliestcpuuse < (N+2)); // sanity check
             assert(earliestcpuuse >= 0); // sanity check
             down_earliest[j] = pj; // was pj+1
             down_latest[j] = earliestcpuuse - 1;  // 181002 added -1
          }
       }
       else { // j is produced on CPU:
          int ngpuconsumers = 0;
          int earliestgpuuse = N+2;
          for (jj=0; jj<ncons[j]; jj++) { // find GPU consumers:
             if (device[cons[j][jj]] == 1) {
                ngpuconsumers++;
                earliestgpuuse = min( earliestgpuuse, cons[j][jj] ); 
             }
          }
          if (ngpuconsumers > 0) { // upload necessary
             assert(earliestgpuuse < (N+2)); // sanity check
             assert(earliestgpuuse >= 0); // sanity check
             up_earliest[j] = pj+1;
             up_latest[j] = earliestgpuuse;
          }
       }
    } // else, keep default values.
 }
 // done.
 // If an _earliest[j]  value remains -1, then j is not used on GPU. 
 // If a _latest[j]  value remains N+2, then j is not computed on GPU. 
 
#if DEBUG
 // test output:
 for (j=0; j<k; j++) {
    printf("  up_earliest[%c] = %d\n", lett(j), up_earliest[j] );
    printf("  up_latest[%c] = %d\n", lett(j), up_latest[j] );
    printf("  down_earliest[%c] = %d\n", lett(j), down_earliest[j] );
    printf("  down_latest[%c] = %d\n\n", lett(j), down_latest[j] );
 }
#endif
 // calculate unary representation in up_possible, down_possible:
 for (i=0; i<N; i++) {
   for (j=0; j<MAX_K; j++) {
     if (up_latest[j] < N+2 && i>=up_earliest[j] && i<=up_latest[j])
       up_possible[i][j] = 1;
     else
       up_possible[i][j] = 0;
   }
 }
 for (i=0; i<=N; i++) {
   for (j=0; j<MAX_K; j++) {
     if (down_earliest[j] > -1 && i>=down_earliest[j] && i<=down_latest[j])
       down_possible[i][j] = 1;
     else
       down_possible[i][j] = 0;
   }
 }
#ifdef DEBUG
 printf("\nup_possible:\n");
 for (i=0; i<N; i++) {
   printf("%3d: ", i);
   for (j=0; j<MAX_K; j++) 
     if (up_possible[i][j])
        printf(" %c", lett(j));
     else
        printf(" .");
   printf("\n");
 }
 printf("\ndown_possible:\n");
 for (i=0; i<=N; i++) {
   printf("%3d: ", i);
   for (j=0; j<MAX_K; j++) 
     if (down_possible[i][j])
        printf(" %c", lett(j));
     else
        printf(" .");
   printf("\n");
 }
#endif
}


/* hamDFS: works on the aff[][] graph and produces i-th hamilton graph.
 * From current node (vector) v, follow unvisited nodes along edges,
 * breaking ties randomly if several edges go to unvisited nodes,
 * but no siblings are followed later. 
 * If a node has no edges at all, mark it visited and return.
 * Uses visited array for marking nodes to stop recursion.
 */
void hamDFS( int v, int i )
{
 int j, jj, jstart;
 //if (visited[v]) return;  // should not happen?
 assert( visited[v] == 0); // sanity check
 visited[v] = 1;
 jstart=rn(0,k-1);  // randomize start index to search for outgoing edges
 if (i==0) { // first search is greedy: start with edge with largest weight
    float maxweight = -1.0;
    int jjstart = jstart;
    for (jj=0; jj<k; jj++) {
       if (jj==v) continue;
       if (aff[v][jj]==0.0) continue;
       if (aff[v][jj] > maxweight) {
          maxweight = aff[v][jj];
          jjstart = jj;
       }
    }
    jstart = jjstart;
    //printf("start with edge to %c\n", lett(jstart));
 }
 for (jj=jstart; jj<jstart+k; jj++) { 
    j = jj%k; // shift into range 0...k-1
    if (j==v) continue;
    if (aff[v][j] > 0) { // edge (v,j) exists:
       if (visited[j])
          continue; // try next j
       else {
          hamDFS( j, i );
          hamilton[i][v][j] = aff[v][j];
          assert( aff[j][v]==aff[v][j]);
          hamilton[i][j][v] = aff[j][v];
          hamlength[i]++;
          weightedhamlength[i]+=aff[v][j];
          //printf("added %d-hamiltonian edge (%c,%c:%f)\n", i, lett(v), lett(j), aff[v][j] ); // test
          break;  // quit j loop, do not consider other sibling edges
       }
    }
 }
}

/* Calculate Hamiltonian paths with randomized heuristic
 * into graph array hamilton[i][][], and return the length.
 */
void comp_hamiltonian( int i )  // i-th attempt -> hamilton[i][][]
{
 int j, jstart, jj;
 // initialize visited array:
 for (j=0; j<k; j++) visited[j] = 0;

 // Starting with a random node, find an unvisited node, and call hamDFS,
 // until all nodes are visited.
 jstart=rn(0,k-1);  // randomize start index to search from
 for (jj=jstart; jj<jstart+k; jj++) { 
    j = jj%k; // shift into range 0...k-1
    if (visited[j])
       continue; 
    hamDFS( j, i );
 }
 // test output:
 //printf(" Hamilton %d, length %d, weightedlength %f:\n", i, hamlength[i], weightedhamlength[i]);
 for (j=0; j<k; j++)
    for (jj=0; jj<k; jj++) {
       if (j==jj) continue;
       if (hamilton[i][j][jj]) {
          assert( aff[j][jj] > 0 ); // sanity check
          //printf( " (%c,%c: %2.1f) ", lett(j), lett(jj), aff[j][jj] );
       }
    }
 //printf("\n\n");
}


/* Calculate Nattempts Hamiltonian paths with randomized heuristic
 * into graph array hamilton[][][], and pick the best one.
 */
void comp_hamiltonians()
{
 int i, j, jj;
 // init:
 for (i=0; i<Nattempts; i++) {
    hamlength[i] = 0;
    weightedhamlength[i] = 0.0;
 }
 for (i=0; i<Nattempts; i++) 
    for (j=0; j<MAX_K; j++) 
       for (jj=0; jj<MAX_K; jj++) 
          hamilton[i][j][jj] = 0;

 for (i=0; i<Nattempts; i++) {
    comp_hamiltonian( i );
 }
 bestweightedhamilton = -1; 
 bestweightedhamlength = -1.0; 
 for (i=0; i<Nattempts; i++) {
    if (weightedhamlength[i] > bestweightedhamlength) {
       bestweightedhamilton = i;
       bestweightedhamlength = weightedhamlength[i];
    }
 }
 printf("Best Hamiltonian is %d, length %d\n", bestweightedhamilton, bestweightedhamlength );
}


/* degree_bh:  determine degree of node in hamilton[besthamilton]
 *             adjacency matrix edge
 */
int degree_bh( int node)
{
 int j;
 int d=0;
 assert(node>=0);
 assert(node<k);
 for (j=0; j<k; j++) {
    if (j==node) continue;
    if (hamilton[bestweightedhamilton][node][j]>0.0) {
       assert(hamilton[bestweightedhamilton][j][node]>0.0);
       d++;
    }
 }
 //printf("degree(%c)=%d\n", lett(node), d); // test
 assert(d<=2);
 return d;
}
    

/* naive code generation - eager placement and allocation, no transfer fusion.
 * fp is NULL (no source code output) or an open CUDA .cu source file.
 */
void emit_naive( FILE *fp )
{
 int i, j, pos=0;
 int uploaded[MAX_K], downloaded[MAX_K];
 // uploaded[j] == 1 iff vector j currently uploaded on device.

 if (!fp) return;  // no screen output for naive code

 // generate Prologue of .cu file:
 fprintf(fp, "#include <cuda.h>\n#include <stdio.h>\n#include <sys/time.h>\n#include <cuda_runtime.h>\n\n");
 fprintf(fp, "#include \"kernels.cu\"\n\n");
 fprintf(fp, "#ifndef NITER\n#define NITER 100\n#endif\n\n");
 fprintf(fp, "#ifndef LOGVLENGTH\n#define LOGVLENGTH 20\n#endif\n#define VLENGTH (1<<LOGVLENGTH)\n\n");
 fprintf(fp, "int main( void )\n{\n");

 for (j=0; j<k; j++) { // allocate vectors (in program order):
    fprintf( fp, " float *h_%c; // vector %c\n h_%c = (float *)malloc( VLENGTH * sizeof(float));\n", lett(j), lett(j), lett(j) );
    fprintf( fp, " float *d_%c; // vector %c\n cudaMalloc((void **)&d_%c, VLENGTH * sizeof(float));\n", lett(j), lett(j), lett(j) );
 }

 // set uploaded, downloaded to 0:
 for (j=0; j<k; j++)  // reset flags for emitting upload/download code:
    uploaded[j] = downloaded[j] = 0; 

 // now all data structures are allocated. Add a timing loop:
 fprintf(fp, " struct timeval nstarttime, nendtime;\n int naivetime;\n gettimeofday( &nstarttime,NULL);\n int iter;\n for (iter=0; iter<NITER; iter++) {\n");

 for (i=0; i<N; i++) { // Kernel calls:

    // check coherence of input operands:
    if (device[i]==1) { // GPU kernel: upload inputs if not there yet:
      for (j=0; j<indeg[i]; j++) {
        if (in[i][j]<INIT_K || device[prod[in[i][j]][0]]==0) {
          if (!uploaded[in[i][j]]) {
            fprintf(fp, "  cudaMemcpy( d_%c, h_%c, VLENGTH*sizeof(float), cudaMemcpyHostToDevice );\n",
                      lett(in[i][j]), lett(in[i][j]) );
            uploaded[in[i][j]] = 1;
          }
        }
      }
    }
    else { // CPU kernel: download inputs if produced on GPU:
      for (j=0; j<indeg[i]; j++) {
        if (in[i][j]>=INIT_K && device[prod[in[i][j]][0]]==1) {
          if (!downloaded[in[i][j]]
            // Bug fixed 190411: In kernel N, need not download those with no CPU consumer
	        && ( device[cons[in[i][j]][0]]==0
		      || ncons[in[i][j]]==2 && device[cons[in[i][j]][1]]==0
		      || ncons[in[i][j]]==3 && device[cons[in[i][j]][2]]==0
		      || ncons[in[i][j]]==4 && device[cons[in[i][j]][3]]==0
		      || ncons[in[i][j]]==5 && device[cons[in[i][j]][4]]==0
		      || ncons[in[i][j]]==6 && device[cons[in[i][j]][5]]==0
		      || ncons[in[i][j]]==7 && device[cons[in[i][j]][6]]==0
		      || ncons[in[i][j]]==8 && device[cons[in[i][j]][7]]==0 // enough
		     ) 
	     )	{
                    fprintf(fp, "  cudaMemcpy( h_%c, d_%c, VLENGTH*sizeof(float), cudaMemcpyDeviceToHost );\n",
                    lett(in[i][j]), lett(in[i][j]) );
            downloaded[in[i][j]] = 1;
          }
        }
      }
    }

    // emit kernel calls:
    if (device[i] == 0) // CPU:
       fprintf(fp, "  CPU%s_%d_%d ( VLENGTH,", name[i], indeg[i], outdeg[i] );
    else // GPU:
       fprintf(fp, "  %s_%d_%d <<< (VLENGTH+255)/256,256>>>( VLENGTH, ", name[i], indeg[i], outdeg[i] );
    // argument list:
    for (j=0; j<indeg[i]; j++) {
       if (device[i]==0)
          fprintf(fp, " h_%c%c", lett(in[i][j]), (outdeg[i]==0&&j==indeg[i]-1)?' ':',');
       else 
          fprintf(fp, " d_%c%c", lett(in[i][j]), (outdeg[i]==0&&j==indeg[i]-1)?' ':',');
    }
    for (j=0; j<outdeg[i]; j++) {
       if (device[i]==0)
          fprintf(fp, " h_%c%c", lett(out[i][j]), j==outdeg[i]-1?' ':',');
       else
          fprintf(fp, " d_%c%c", lett(out[i][j]), j==outdeg[i]-1?' ':',');
    }
    if (device[i] == 0) {
       if (fp) fprintf( fp, "); /*CPU Kernel call %d*/\n", i );
    }
    else {
       if (fp) fprintf( fp, " ); /*GPU Kernel call %d*/\n", i );
    }
 }

 // pending downloads of live-on-exit vectors after N-1:
 for (j=INIT_K; j<k; j++) {
   if (device[prod[j][0]]==1 && !downloaded[j]
      //(down_possible[N][j] && !downloaded[j]) 
      // Bug fixed 190411: need only download if any consumer on GPU
        && ( device[cons[j][0]]==0 
             || ncons[j]>1 && device[cons[j][1]]==0 
             || ncons[j]>2 && device[cons[j][2]]==0 
             || ncons[j]>3 && device[cons[j][3]]==0 
             || ncons[j]>4 && device[cons[j][4]]==0 
             || ncons[j]>5 && device[cons[j][5]]==0 
             || ncons[j]>6 && device[cons[j][6]]==0 
             || ncons[j]>7 && device[cons[j][7]]==0 
             || ncons[j]>8 && device[cons[j][8]]==0  )  // enough
      )
   {
      fprintf( fp, "  cudaMemcpy( h_%c, d_%c, VLENGTH*sizeof(float), cudaMemcpyDeviceToHost );\n",
               lett(j), lett(j) );
      downloaded[j] = 1;
   }
 }

 // epilogue, including timing loop:
      fprintf(fp, " }\n gettimeofday( &nendtime, NULL);\n naivetime=(nendtime.tv_sec-nstarttime.tv_sec)*1000000 + (nendtime.tv_usec-nstarttime.tv_usec);\n printf(\" Time (avgd. over %%d iter.) with naive transfers: %%d us\\n\", NITER, naivetime/NITER);\n}\n");
}


/* naive code generation with memory allocation included in the measurement loop
 * - eager placement and allocation, no transfer fusion.
 * fp is NULL (no source code output) or an open CUDA .cu source file.
 */
void emit_naive_alloc( FILE *fp )
{
 int i, j, pos=0;
 int uploaded[MAX_K], downloaded[MAX_K];
 // uploaded[j] == 1 iff vector j currently uploaded on device.

 if (!fp) return;  // no screen output for naive code

 // generate Prologue of .cu file:
 fprintf(fp, "#include <cuda.h>\n#include <stdio.h>\n#include <sys/time.h>\n#include <cuda_runtime.h>\n\n");
 fprintf(fp, "#include \"kernels.cu\"\n\n");
 fprintf(fp, "#ifndef NITER\n#define NITER 100\n#endif\n\n");
 fprintf(fp, "#ifndef LOGVLENGTH\n#define LOGVLENGTH 20\n#endif\n#define VLENGTH (1<<LOGVLENGTH)\n\n");
 fprintf(fp, "int main( void )\n{\n");

 // timing loop spans everything:
 fprintf(fp, " struct timeval nstarttime, nendtime;\n int naivetime;\n float *d_dummy; cudaMalloc((void **)&d_dummy, sizeof(float));\n gettimeofday( &nstarttime,NULL);\n int iter;\n for (iter=0; iter<NITER; iter++) {\n");

 for (j=0; j<k; j++) { // allocate vectors (in program order):
    fprintf( fp, " float *h_%c; // vector %c\n h_%c = (float *)malloc( VLENGTH * sizeof(float));\n", lett(j), lett(j), lett(j) );
    fprintf( fp, " float *d_%c; // vector %c\n cudaMalloc((void **)&d_%c, VLENGTH * sizeof(float));\n", lett(j), lett(j), lett(j) );
 }

 // set uploaded, downloaded to 0:
 for (j=0; j<k; j++)  // reset flags for emitting upload/download code:
    uploaded[j] = downloaded[j] = 0; 

 for (i=0; i<N; i++) { // Kernel calls:

    // check coherence of input operands:
    if (device[i]==1) { // GPU kernel: upload inputs if not there yet:
      for (j=0; j<indeg[i]; j++) {
        if (in[i][j]<INIT_K || device[prod[in[i][j]][0]]==0) {
          if (!uploaded[in[i][j]]) {
            fprintf(fp, "  cudaMemcpy( d_%c, h_%c, VLENGTH*sizeof(float), cudaMemcpyHostToDevice );\n",
                      lett(in[i][j]), lett(in[i][j]) );
            uploaded[in[i][j]] = 1;
          }
        }
      }
    }
    else { // CPU kernel: download inputs if produced on GPU:
      for (j=0; j<indeg[i]; j++) {
        if (in[i][j]>=INIT_K && device[prod[in[i][j]][0]]==1) {
          if (!downloaded[in[i][j]]
            // Bug fixed 190411: In kernel N, need not download those with no CPU consumer
	        && ( device[cons[in[i][j]][0]]==0
		      || ncons[in[i][j]]==2 && device[cons[in[i][j]][1]]==0
		      || ncons[in[i][j]]==3 && device[cons[in[i][j]][2]]==0
		      || ncons[in[i][j]]==4 && device[cons[in[i][j]][3]]==0
		      || ncons[in[i][j]]==5 && device[cons[in[i][j]][4]]==0
		      || ncons[in[i][j]]==6 && device[cons[in[i][j]][5]]==0
		      || ncons[in[i][j]]==7 && device[cons[in[i][j]][6]]==0
		      || ncons[in[i][j]]==8 && device[cons[in[i][j]][7]]==0 // enough
		     ) 
	     )	{
                    fprintf(fp, "  cudaMemcpy( h_%c, d_%c, VLENGTH*sizeof(float), cudaMemcpyDeviceToHost );\n",
                    lett(in[i][j]), lett(in[i][j]) );
            downloaded[in[i][j]] = 1;
          }
        }
      }
    }

    // emit kernel calls:
    if (device[i] == 0) // CPU:
       fprintf(fp, "  CPU%s_%d_%d ( VLENGTH,", name[i], indeg[i], outdeg[i] );
    else // GPU:
       fprintf(fp, "  %s_%d_%d <<< (VLENGTH+255)/256,256>>>( VLENGTH, ", name[i], indeg[i], outdeg[i] );
    // argument list:
    for (j=0; j<indeg[i]; j++) {
       if (device[i]==0)
          fprintf(fp, " h_%c%c", lett(in[i][j]), (outdeg[i]==0&&j==indeg[i]-1)?' ':',');
       else 
          fprintf(fp, " d_%c%c", lett(in[i][j]), (outdeg[i]==0&&j==indeg[i]-1)?' ':',');
    }
    for (j=0; j<outdeg[i]; j++) {
       if (device[i]==0)
          fprintf(fp, " h_%c%c", lett(out[i][j]), j==outdeg[i]-1?' ':',');
       else
          fprintf(fp, " d_%c%c", lett(out[i][j]), j==outdeg[i]-1?' ':',');
    }
    if (device[i] == 0) {
       if (fp) fprintf( fp, "); /*CPU Kernel call %d*/\n", i );
    }
    else {
       if (fp) fprintf( fp, " ); /*GPU Kernel call %d*/\n", i );
    }
 }

 // pending downloads of live-on-exit vectors after N-1:
 for (j=INIT_K; j<k; j++) {
   if (device[prod[j][0]]==1 && !downloaded[j]
      //(down_possible[N][j] && !downloaded[j]) 
      // Bug fixed 190411: need only download if any consumer on GPU
        && ( device[cons[j][0]]==0 
             || ncons[j]>1 && device[cons[j][1]]==0 
             || ncons[j]>2 && device[cons[j][2]]==0 
             || ncons[j]>3 && device[cons[j][3]]==0 
             || ncons[j]>4 && device[cons[j][4]]==0 
             || ncons[j]>5 && device[cons[j][5]]==0 
             || ncons[j]>6 && device[cons[j][6]]==0 
             || ncons[j]>7 && device[cons[j][7]]==0 
             || ncons[j]>8 && device[cons[j][8]]==0  )  // enough
      )
   {
      fprintf( fp, "  cudaMemcpy( h_%c, d_%c, VLENGTH*sizeof(float), cudaMemcpyDeviceToHost );\n",
               lett(j), lett(j) );
      downloaded[j] = 1;
   }
 }

 // deallocate all vectors on host and device: 
 for (j=0; j<k; j++) { // (in program order):
    fprintf( fp, " free(h_%c);\n", lett(j) );
    fprintf( fp, " cudaFree(d_%c);\n", lett(j) );
 }
 // epilogue, including timing loop:
      fprintf(fp, " }\n gettimeofday( &nendtime, NULL);\n naivetime=(nendtime.tv_sec-nstarttime.tv_sec)*1000000 + (nendtime.tv_usec-nstarttime.tv_usec);\n printf(\" Time (avgd. over %%d iter.) with naive transfers: %%d us\\n\", NITER, naivetime/NITER);\n}\n");
}



/* fp is NULL (no source code output) or an open CUDA .cu source file.
 */
void emit_code( FILE *fp )
{
 int i, j, pos=0;
 int uploaded[MAX_K], downloaded[MAX_K];
 // uploaded[j] == 1 iff vector j currently uploaded on device.

 if (fp) { // generate Prologue of .cu file:
    fprintf(fp, "#include <cuda.h>\n#include <stdio.h>\n#include <sys/time.h>\n#include <cuda_runtime.h>\n\n");
    fprintf(fp, "#include \"kernels.cu\"\n\n");
    fprintf(fp, "#ifndef NITER\n#define NITER 100\n#endif\n\n");
    fprintf(fp, "#ifndef LOGVLENGTH\n#define LOGVLENGTH 20\n#endif\n#define VLENGTH (1<<LOGVLENGTH)\n\n");
    fprintf(fp, "int main( void )\n{\n");
 }

 for (j=0; j<k; j++)
    visited[j] = 0;  // needed for traversal 

 for (j=0; j<k; j++) { // emit all vectors:
    int deg, jjj, jjjj, found=0;
    int toallocate_j=-1; 
    int toallocate_pos = -1;
    int toallocate_vectors = -1;

    if (visited[j]) continue; // already emitted
    deg = degree_bh(j);

    if (deg==0) {
       visited[j]=1;
       position[j] = pos;
       rev_position[pos] = j;
       //printf(" vector %c (singleton) at position %d\n", lett(j), pos++ );
       if (fp) {
          fprintf( fp, " float *h_%c; // vector %c (singleton) alloc at position %d\n h_%c = (float *)malloc( VLENGTH * sizeof(float));\n", lett(j), lett(j), pos, lett(j) );
          fprintf( fp, " float *d_%c; // vector %c (singleton) alloc at position %d\n cudaMalloc((void **)&d_%c, VLENGTH * sizeof(float));\n", lett(j), lett(j), pos, lett(j) );
       }
       pos++;
       continue;
    }

    if (deg==2)
       continue;  // binary nodes will be visited by path from a degree-1 node

    assert(deg==1);
    found = 1;
    // found an unvisited unary node j in bh to start with:
    jjj=j;
    do {
       position[jjj] = pos;
       rev_position[pos] = jjj;
       visited[jjj] = 1;
       //printf(" vector %c at position %d\n", lett(jjj), pos );
       if (fp) {
          if (jjj==j) { // first in Hamiltonian of this CC:
             fprintf( fp, " float *h_%c, *d_%c; // vector %c alloc at position %d\n",
                      lett(j), lett(j), lett(j), pos );
             // delay alloc code generation for entire Hamiltonian:
             assert( toallocate_vectors == -1);
             toallocate_j=j; toallocate_pos = pos; toallocate_vectors = 1;
          }
          else {
             fprintf( fp, " float *h_%c, *d_%c; // vector %c cont. at position %d\n",
                      lett(jjj), lett(jjj), lett(jjj), pos );
             assert( toallocate_vectors > 0);
             toallocate_vectors++;
          }
       }
       pos++;
       // find next in path:
       found=0; 
       for (jjjj=0; jjjj<k; jjjj++) { // find continuation
          if (visited[jjjj]) continue;
          if (hamilton[bestweightedhamilton][jjj][jjjj]>0.0) {
             assert (hamilton[bestweightedhamilton][jjjj][jjj]>0.0);
             jjj = jjjj;
             found=1;
             break; // found new jjj
          }
       }
       // leave whileloop if not found:
    } while (found); // while (degree_bh(jjj)==2);
    // now go to the last one in the Hamiltonian:
    if (pos>k) assert(0); // upper limit for all found, should not occur
    // now emit allocation code for entire Hamiltonian:
    assert(toallocate_vectors > 0);
    if (fp) {
       int cnt, q;
       fprintf(fp, " h_%c = (float *)malloc( %d * VLENGTH * sizeof(float));\n",
               lett(toallocate_j), toallocate_vectors ); 
       fprintf(fp, " cudaMalloc((void **)&d_%c, %d * VLENGTH * sizeof(float));\n",
               lett(toallocate_j), toallocate_vectors ); 
       for (cnt = 1; cnt < toallocate_vectors; cnt++) {
           q = rev_position[toallocate_pos+cnt];
           fprintf(fp, " h_%c = h_%c + %d * VLENGTH;\n",
                   lett(q), lett(toallocate_j), cnt );
           fprintf(fp, " d_%c = d_%c + %d * VLENGTH;\n",
                   lett(q), lett(toallocate_j), cnt );
       }
    }
    toallocate_j = toallocate_pos = toallocate_vectors = -1;
 }

 // now all data structures are allocated. Add a timing loop:
 if (fp)
    fprintf(fp, " struct timeval fstarttime, fendtime;\n int fusedtime;\n gettimeofday( &fstarttime,NULL);\n int iter;\n for (iter=0; iter<NITER; iter++) {\n");

 // determine upload and download times and ordering:
 for (j=0; j<k; j++) {
    uploaded[j] = downloaded[j] = 0;
 }
 for (i=0; i<N; i++) { 
    uploads_from[i] = -1;  // init
    uploads_until[i] = -1;
 }
 for (i=0; i<N; i++) { // go through up_possible:
#if DEBUG
    printf("   ... step %d\n", i );
#endif
    for (j=0; j<k; j++) { // go through memory positions 0,...
       int prevjjjj, prevprevjjjj, jj, jjj, ci=0, cipp=0;
       jj = rev_position[j]; // the vector in memory position j 
       if (!up_possible[i][jj]) continue; 
       if (uploaded[jj]) continue; // nothing more to do for j
       //printf("        ... test pos j=%d, vector jj=%c\n", j, lett(jj));
       // find ci = # vectors jjjj following jj in memory 
       //     for which up_possible is also 1 at time i
       jjj=j+1;
       prevjjjj = jj;
       prevprevjjjj = jj; 
       ci=1;
       while (jjj<k) { // following memory positions after j:
          int jjjj = rev_position[jjj]; // vector at jjj
          if (!up_possible[i][jjjj]) break;
          if (uploaded[jjjj]) break;
          // also, break if there is no affinity edge from previous jjjj to this one.
          if (!aff[prevjjjj][jjjj]) break;
          // heuristic pruning condition: (test 181002:)
#ifdef TIGHT
          if (prevjjjj != prevprevjjjj)
            if (aff[prevjjjj][jjjj] < aff[prevprevjjjj][prevjjjj]) break;
#endif
          // tentatively include jjjj:
          ci++; 
          jjj++;
          prevprevjjjj=prevjjjj; 
          prevjjjj=jjjj; 
       }
       //printf("            ... ci=%d\n", ci ); 
       if (i<N-1 && up_possible[i+1][jj]) {
        // find cipp = # vectors jjjj following jj in memory 
        //     for which up_possible is also 1 at time i+1
        jjj=j+1;
        prevjjjj = jj;
        cipp=1;
        while (jjj<k) { 
          int jjjj = rev_position[jjj];
          if (!up_possible[i+1][jjjj]) break;
          if (uploaded[jjjj]) break;
          // also, break if there is no affinity edge from previous jjjj to this one.
          if (!aff[prevjjjj][jjjj]) break;
#ifdef TIGHT
          if (prevjjjj != prevprevjjjj)
            if (aff[prevjjjj][jjjj] < aff[prevprevjjjj][prevjjjj]) break;
#endif
          cipp++; 
          jjj++;
          prevjjjj=jjjj; 
        }
        //printf("            ... cipp=%d\n", cipp ); 
       }
       // greedy strategy to find best co-uploads:
       if (ci > cipp || i==N-1) { // upload pos j...j+ci at time i
          //printf("             ... upload from pos %d to %d at time %d\n", j, j+ci-1, i);
          assert(ci>0);
          if (uploads_from[i]==-1)
             uploads_from[i] = j;   // otherwise keep earlier lower bound
          uploads_until[i] = j+ci-1;
          for (jjj=j; jjj<j+ci; jjj++) {
             //printf("           %d -> %c\n", jjj, lett(rev_position[jjj]) );
             uploaded[rev_position[jjj]] = 1; // mark as done for upload
          }
       }
    }
 }

 // now find the downloads:

 for (i=0; i<N; i++) { 
    downloads_from[i] = -1;  // init
    downloads_until[i] = -1;
 }
 for (i=0; i<=N; i++) { // go through down_possible:
#if DEBUG
    printf("   ... find downloads, step %d\n", i );
#endif
    for (j=0; j<k; j++) { // go through memory positions 0,...
       int prevjjjj, prevprevjjjj, jj, jjj, ci=0, cipp=0;
       jj = rev_position[j]; // the vector in memory position j 
       if (!down_possible[i][jj]) continue; 
       if (downloaded[jj]) continue; // nothing more to do for j
       //printf("        ... test pos j=%d, vector jj=%c\n", j, lett(jj));
       // find ci = # vectors jjjj following jj in memory 
       //     for which down_possible is also 1 at time i
       jjj=j+1;
       prevjjjj = jj;
       prevprevjjjj = jj; 
       ci=1;
       while (jjj<k) { // following memory positions after j:
          int jjjj = rev_position[jjj]; // vector at jjj
          if (!down_possible[i][jjjj]) break;
          if (downloaded[jjjj]) break;
          // also, break if there is no affinity edge from previous jjjj to this one.
          if (!aff[prevjjjj][jjjj]) break;
          // download as early as possible, i.e. no heuristic break
          // tentatively include jjjj:
          ci++; 
          jjj++;
          prevprevjjjj=prevjjjj; 
          prevjjjj=jjjj; 
       }
       //printf("            ... ci=%d\n", ci ); 
       if (i<N-1 && down_possible[i+1][jj]) {
        // find cipp = # vectors jjjj following jj in memory 
        //     for which down_possible is also 1 at time i+1
        jjj=j+1;
        prevjjjj = jj;
        cipp=1;
        while (jjj<k) { 
          int jjjj = rev_position[jjj];
          if (!down_possible[i+1][jjjj]) break;
          if (downloaded[jjjj]) break;
          // also, break if there is no affinity edge from previous jjjj to this one.
          if (!aff[prevjjjj][jjjj]) break;
          cipp++; 
          jjj++;
          prevjjjj=jjjj; 
        }
        //printf("            ... cipp=%d\n", cipp ); 
       }
       // greedy strategy to find best co-downloads:
       if (i>=N-1 || ci > cipp) { // download pos j...j+ci at time i
          //printf("             ... download from pos %d to %d at time %d\n", j, j+ci-1, i);
          assert(ci>0);
          if (downloads_from[i]==-1)
             downloads_from[i] = j;   // otherwise keep earlier lower bound
          downloads_until[i] = j+ci-1;
          for (jjj=j; jjj<j+ci; jjj++) {
             //printf("           %d -> %c\n", jjj, lett(rev_position[jjj]) );
             downloaded[rev_position[jjj]] = 1; // mark as done for download
          }
       }
    }
 }

 // could later insert some checking code here to make sure that
 // all containers with a consumer on GPU are uploaded on time
 // and ... with a producer on GPU are downloaded on time...

 for (j=0; j<k; j++)  // reset flags for emitting upload/download code:
    uploaded[j] = downloaded[j] = 0; 

 savings = 0; // record number of fused messages

 gettimeofday( &printstarttime, NULL);  // take the time 181119

 for (i=0; i<N; i++) { // emit kernels:
    // bulk uploads before i:
    if (uploads_from[i] > -1) { // bulk uploads to do (first):
       for (j=uploads_from[i]; j<=uploads_until[i]; j++) {
          assert(uploaded[rev_position[j]] == 0);
          printf(" before %d: %s-upload %c from location %d\n", 
                 i, ((uploads_until[i]-uploads_from[i])>1)?"bulk":"single",
                 lett(rev_position[j]), j );
          uploaded[rev_position[j]] = 1;
       }
       savings += uploads_until[i]-uploads_from[i];
       if (fp) {
          fprintf(fp, "  cudaMemcpy( d_%c, h_%c, %d*VLENGTH*sizeof(float), cudaMemcpyHostToDevice );\n",
                  lett(rev_position[uploads_from[i]]),
                  lett(rev_position[uploads_from[i]]),
                  uploads_until[i]-uploads_from[i]+1 );
       }
    }
    if (device[i]==1) { // make sure that all operands of i are there:
      for (j=0; j<indeg[i]; j++) {
        if (in[i][j]<INIT_K || device[prod[in[i][j]][0]]==0) {
          if (!uploaded[in[i][j]]) {
            printf(" before %d: single-upload %c from location %d\n", 
                   i, lett(in[i][j]), position[in[i][j]] );
            uploaded[in[i][j]] = 1;
            if (fp) {
              fprintf(fp, "  cudaMemcpy( d_%c, h_%c, VLENGTH*sizeof(float), cudaMemcpyHostToDevice );\n",
                      lett(rev_position[j]), lett(rev_position[j]) );
            }
          }
        }
      }
    }
#if 0
    // old code, to be revised:
    for (j=0; j<indeg[i]; j++) { // all input operands
       int jj, jjj, jjjj;
       if (!uploaded[in[i][j]]) {
          int vj = in[i][j];
          // vj must be loaded before i. Not necessarily first.
          // now go through other non-uploaded vectors 
          // that overlap in their earliest_up...latest_up interval
          // and are connected to j in bh.
          // finally, emit in position[] order.
          for (jj=position[vj]; jj>=0; jj--) { // search to left in position
            int v = rev_position[jj];
            if (uploaded[v]) break;
            //... if no overlap, break; TODO
          }
          for (jjj=position[vj]; jjj<k; jjj++) { // search to right in position
            int v = rev_position[jj];
            if (uploaded[v]) break;
            //... if no overlap, break; TODO
            // ... TODO ...
          }
          // candidates jj...j...jjj can be uploaded in this order:
          for (jjjj=jj; jjjj<=jjj; jjjj++) { 
            int v = rev_position[jjjj];
            printf("upload %c\n", lett(jjjj)); // test
            uploaded[v] = 1;
          }
       }
    }
    // end old code to be revised or dropped
#endif

    printf("  Kernel %d (%s) ( " /*)*/, i, name[i] );
    if (fp)
       if (device[i] == 0) // CPU:
          fprintf(fp, "  CPU%s_%d_%d ( VLENGTH,", name[i], indeg[i], outdeg[i] );
       else // GPU:
          fprintf(fp, "  %s_%d_%d <<< (VLENGTH+255)/256,256>>>( VLENGTH, ", name[i], indeg[i], outdeg[i] );
    // argument list:
    for (j=0; j<indeg[i]; j++) {
       printf(" R(%c),", lett(in[i][j]));
       if (fp) 
         if (device[i]==0)
           fprintf(fp, " h_%c%c", lett(in[i][j]), (outdeg[i]==0&&j==indeg[i]-1)?' ':',');
         else 
           fprintf(fp, " d_%c%c", lett(in[i][j]), (outdeg[i]==0&&j==indeg[i]-1)?' ':',');
    }
    for (j=0; j<outdeg[i]; j++) {
       printf(" W(%c)%c", lett(out[i][j]), j==outdeg[i]-1?' ':',');
       if (fp)
         if (device[i]==0)
           fprintf(fp, " h_%c%c", lett(out[i][j]), j==outdeg[i]-1?' ':',');
         else
           fprintf(fp, " d_%c%c", lett(out[i][j]), j==outdeg[i]-1?' ':',');
    }
    if (device[i] == 0) {
       printf(" ) on CPU\n");
       if (fp) fprintf( fp, "); /*CPU Kernel call %d*/\n", i );
    }
    else {
       printf(" ) on GPU\n");
       if (fp) fprintf( fp, " ); /*GPU Kernel call %d*/\n", i );
    }

    // bulk-downloads after i:
    if (downloads_from[i] > -1) { // bulk downloads to do (first):
       for (j=downloads_from[i]; j<=downloads_until[i]; j++) {
          if( downloaded[rev_position[j]] != 0 )
            printf("??? i=%d j=%d vj=%c dnl=%d\n", i, j, lett(rev_position[j]),
                     downloaded[rev_position[j]] );
// BUG: FIXME - assert now eliminated to get through as it only occurs for the last kernel call:
          //assert( downloaded[rev_position[j]] == 0 );
          printf(" after %d: bulk-download %c to location %d\n", 
                 i, lett(rev_position[j]), j );
          downloaded[rev_position[j]] = 1;
       }
       savings += downloads_until[i]-downloads_from[i];
       if (fp)
          fprintf( fp, "  cudaMemcpy( h_%c, d_%c, %d*VLENGTH*sizeof(float), cudaMemcpyDeviceToHost );\n",
                   lett(rev_position[downloads_from[i]]),
                   lett(rev_position[downloads_from[i]]),
                   downloads_until[i]-downloads_from[i]+1 );
    }
#if 0 
    // we do not require all produced values to be downloaded immediately... 
    if (device[i]==1) 
    for (j=0; j<outdeg[i]; j++) {
       if (!downloaded[out[i][j]]) {
          printf(" after %d:  download %c to location %d\n", 
                 i, lett(out[i][j]), position[out[i][j]] );
          downloaded[out[i][j]] = 1;
       }
    }
#endif
 } // for i - loop over kernel calls
#if 0
 // it seems the following is not necessary and only leads to trouble:

 // finally, downloads after N-1 for live on exit vectors:
 if (downloads_from[N] > -1) { // bulk downloads to do (first):
   for (j=downloads_from[N]; j<=downloads_until[N]; j++) {
      printf("HALLO! j=%d\n", j);
      if( downloaded[rev_position[j]] != 0 )
        printf("??? i=N=%d j=%d vj=%c dnl=%d\n", N, j, lett(rev_position[j]),
                 downloaded[rev_position[j]] );
      assert( downloaded[rev_position[j]] == 0 );
      printf(" after %d: download %c to location %d\n", 
             N, lett(rev_position[j]), j );
      downloaded[rev_position[j]] = 1;
   }
   savings += downloads_until[N]-downloads_from[N];
 }
#endif
   // pending downloads of live-on-exit vectors:
   for (j=INIT_K; j<k; j++) {
     if //(down_possible[N][j] && !downloaded[j]) 
        // Fixed bug 190411: need download only if any consumer is on CPU:
        (device[prod[j][0]]==1 && !downloaded[j]
	        && (   device[cons[j][0]]==0
		       || ncons[j]>=2 && device[cons[j][1]]==0
		       || ncons[j]>=3 && device[cons[j][2]]==0
		       || ncons[j]>=4 && device[cons[j][3]]==0
		       || ncons[j]>=5 && device[cons[j][4]]==0
		       || ncons[j]>=6 && device[cons[j][5]]==0
		       || ncons[j]>=7 && device[cons[j][6]]==0
		       || ncons[j]>=8 && device[cons[j][7]]==0  // enough...
		   ) 
		)
     {
        printf(" after %d:   download %c to location %d\n", N, lett(j), position[j] );
        downloaded[j] = 1;
        if (fp) fprintf( fp, "  cudaMemcpy( h_%c, d_%c, VLENGTH*sizeof(float), cudaMemcpyDeviceToHost );\n",
                         lett(j), lett(j) );
     }
   }
   // no savings here

   if (fp) { // epilogue, including timing loop:
      fprintf(fp, " }\n gettimeofday( &fendtime, NULL);\n fusedtime=(fendtime.tv_sec-fstarttime.tv_sec)*1000000 + (fendtime.tv_usec-fstarttime.tv_usec);\n printf(\" Time (avgd. over %%d iter.) with transfer fusion: %%d us\\n\", NITER, fusedtime/NITER);\n}\n");
   }
}


/* Clone of the above, with the memory allocation time included in the measurement.
/* fp is NULL (no source code output) or an open CUDA .cu source file.
 */
void emit_code_alloc( FILE *fp )
{
 int i, j, pos=0;
 int uploaded[MAX_K], downloaded[MAX_K];
 // uploaded[j] == 1 iff vector j currently uploaded on device.

 if (fp) { // generate Prologue of .cu file:
    fprintf(fp, "#include <cuda.h>\n#include <stdio.h>\n#include <sys/time.h>\n#include <cuda_runtime.h>\n\n");
    fprintf(fp, "#include \"kernels.cu\"\n\n");
    fprintf(fp, "#ifndef NITER\n#define NITER 100\n#endif\n\n");
    fprintf(fp, "#ifndef LOGVLENGTH\n#define LOGVLENGTH 20\n#endif\n#define VLENGTH (1<<LOGVLENGTH)\n\n");
    fprintf(fp, "int main( void )\n{\n");
 }
 // Add a timing loop:
 if (fp)
    fprintf(fp, " struct timeval fstarttime, fendtime;\n int fusedtime;\n float *d_dummy; cudaMalloc((void **)&d_dummy,sizeof(float));\n gettimeofday( &fstarttime,NULL);\n int iter;\n for (iter=0; iter<NITER; iter++) {\n");

 for (j=0; j<k; j++) {
    visited[j] = 0;  // needed for traversal 
    calledmalloc[j] = 0;
 }

 for (j=0; j<k; j++) { // emit all vectors:
    int deg, jjj, jjjj, found=0;
    int toallocate_j=-1; 
    int toallocate_pos = -1;
    int toallocate_vectors = -1;

    if (visited[j]) continue; // already emitted
    deg = degree_bh(j);

    if (deg==0) {
       visited[j]=1;
       position[j] = pos;
       rev_position[pos] = j;
       //printf(" vector %c (singleton) at position %d\n", lett(j), pos++ );
       if (fp) {
          fprintf( fp, " float *h_%c; // vector %c (singleton) alloc at position %d\n h_%c = (float *)malloc( VLENGTH * sizeof(float));\n", lett(j), lett(j), pos, lett(j) );
          fprintf( fp, " float *d_%c; // vector %c (singleton) alloc at position %d\n cudaMalloc((void **)&d_%c, VLENGTH * sizeof(float));\n", lett(j), lett(j), pos, lett(j) );
	  calledmalloc[j] = 1; // need to call free later
       }
       pos++;
       continue;
    }

    if (deg==2)
       continue;  // binary nodes will be visited by path from a degree-1 node

    assert(deg==1);
    found = 1;
    // found an unvisited unary node j in bh to start with:
    jjj=j;
    do {
       position[jjj] = pos;
       rev_position[pos] = jjj;
       visited[jjj] = 1;
       //printf(" vector %c at position %d\n", lett(jjj), pos );
       if (fp) {
          if (jjj==j) { // first in Hamiltonian of this CC:
             fprintf( fp, " float *h_%c, *d_%c; // vector %c alloc at position %d\n",
                      lett(j), lett(j), lett(j), pos );
             // delay alloc code generation for entire Hamiltonian:
             assert( toallocate_vectors == -1);
             toallocate_j=j; toallocate_pos = pos; toallocate_vectors = 1;
          }
          else {
             fprintf( fp, " float *h_%c, *d_%c; // vector %c cont. at position %d\n",
                      lett(jjj), lett(jjj), lett(jjj), pos );
             assert( toallocate_vectors > 0);
             toallocate_vectors++;
          }
       }
       pos++;
       // find next in path:
       found=0; 
       for (jjjj=0; jjjj<k; jjjj++) { // find continuation
          if (visited[jjjj]) continue;
          if (hamilton[bestweightedhamilton][jjj][jjjj]>0.0) {
             assert (hamilton[bestweightedhamilton][jjjj][jjj]>0.0);
             jjj = jjjj;
             found=1;
             break; // found new jjj
          }
       }
       // leave whileloop if not found:
    } while (found); // while (degree_bh(jjj)==2);
    // now go to the last one in the Hamiltonian:
    if (pos>k) assert(0); // upper limit for all found, should not occur
    // now emit allocation code for entire Hamiltonian:
    assert(toallocate_vectors > 0);
    if (fp) {
       int cnt, q;
       fprintf(fp, " h_%c = (float *)malloc( %d * VLENGTH * sizeof(float));\n",
               lett(toallocate_j), toallocate_vectors ); 
       fprintf(fp, " cudaMalloc((void **)&d_%c, %d * VLENGTH * sizeof(float));\n",
               lett(toallocate_j), toallocate_vectors ); 
       calledmalloc[toallocate_j] = 1; // need to call free later
       for (cnt = 1; cnt < toallocate_vectors; cnt++) {
           q = rev_position[toallocate_pos+cnt];
           fprintf(fp, " h_%c = h_%c + %d * VLENGTH;\n",
                   lett(q), lett(toallocate_j), cnt );
           fprintf(fp, " d_%c = d_%c + %d * VLENGTH;\n",
                   lett(q), lett(toallocate_j), cnt );
       }
    }
    toallocate_j = toallocate_pos = toallocate_vectors = -1;
 }

 // now all data structures are allocated.

 // determine upload and download times and ordering:
 for (j=0; j<k; j++) {
    uploaded[j] = downloaded[j] = 0;
 }
 for (i=0; i<N; i++) { 
    uploads_from[i] = -1;  // init
    uploads_until[i] = -1;
 }
 for (i=0; i<N; i++) { // go through up_possible:
#if DEBUG
    printf("   ... step %d\n", i );
#endif
    for (j=0; j<k; j++) { // go through memory positions 0,...
       int prevjjjj, prevprevjjjj, jj, jjj, ci=0, cipp=0;
       jj = rev_position[j]; // the vector in memory position j 
       if (!up_possible[i][jj]) continue; 
       if (uploaded[jj]) continue; // nothing more to do for j
       //printf("        ... test pos j=%d, vector jj=%c\n", j, lett(jj));
       // find ci = # vectors jjjj following jj in memory 
       //     for which up_possible is also 1 at time i
       jjj=j+1;
       prevjjjj = jj;
       prevprevjjjj = jj; 
       ci=1;
       while (jjj<k) { // following memory positions after j:
          int jjjj = rev_position[jjj]; // vector at jjj
          if (!up_possible[i][jjjj]) break;
          if (uploaded[jjjj]) break;
          // also, break if there is no affinity edge from previous jjjj to this one.
          if (!aff[prevjjjj][jjjj]) break;
          // heuristic pruning condition: (test 181002:)
#ifdef TIGHT
          if (prevjjjj != prevprevjjjj)
            if (aff[prevjjjj][jjjj] < aff[prevprevjjjj][prevjjjj]) break;
#endif
          // tentatively include jjjj:
          ci++; 
          jjj++;
          prevprevjjjj=prevjjjj; 
          prevjjjj=jjjj; 
       }
       //printf("            ... ci=%d\n", ci ); 
       if (i<N-1 && up_possible[i+1][jj]) {
        // find cipp = # vectors jjjj following jj in memory 
        //     for which up_possible is also 1 at time i+1
        jjj=j+1;
        prevjjjj = jj;
        cipp=1;
        while (jjj<k) { 
          int jjjj = rev_position[jjj];
          if (!up_possible[i+1][jjjj]) break;
          if (uploaded[jjjj]) break;
          // also, break if there is no affinity edge from previous jjjj to this one.
          if (!aff[prevjjjj][jjjj]) break;
#ifdef TIGHT
          if (prevjjjj != prevprevjjjj)
            if (aff[prevjjjj][jjjj] < aff[prevprevjjjj][prevjjjj]) break;
#endif
          cipp++; 
          jjj++;
          prevjjjj=jjjj; 
        }
        //printf("            ... cipp=%d\n", cipp ); 
       }
       // greedy strategy to find best co-uploads:
       if (ci > cipp || i==N-1) { // upload pos j...j+ci at time i
          //printf("             ... upload from pos %d to %d at time %d\n", j, j+ci-1, i);
          assert(ci>0);
          if (uploads_from[i]==-1)
             uploads_from[i] = j;   // otherwise keep earlier lower bound
          uploads_until[i] = j+ci-1;
          for (jjj=j; jjj<j+ci; jjj++) {
             //printf("           %d -> %c\n", jjj, lett(rev_position[jjj]) );
             uploaded[rev_position[jjj]] = 1; // mark as done for upload
          }
       }
    }
 }

 // now find the downloads:

 for (i=0; i<N; i++) { 
    downloads_from[i] = -1;  // init
    downloads_until[i] = -1;
 }
 for (i=0; i<=N; i++) { // go through down_possible:
#if DEBUG
    printf("   ... find downloads, step %d\n", i );
#endif
    for (j=0; j<k; j++) { // go through memory positions 0,...
       int prevjjjj, prevprevjjjj, jj, jjj, ci=0, cipp=0;
       jj = rev_position[j]; // the vector in memory position j 
       if (!down_possible[i][jj]) continue; 
       if (downloaded[jj]) continue; // nothing more to do for j
       //printf("        ... test pos j=%d, vector jj=%c\n", j, lett(jj));
       // find ci = # vectors jjjj following jj in memory 
       //     for which down_possible is also 1 at time i
       jjj=j+1;
       prevjjjj = jj;
       prevprevjjjj = jj; 
       ci=1;
       while (jjj<k) { // following memory positions after j:
          int jjjj = rev_position[jjj]; // vector at jjj
          if (!down_possible[i][jjjj]) break;
          if (downloaded[jjjj]) break;
          // also, break if there is no affinity edge from previous jjjj to this one.
          if (!aff[prevjjjj][jjjj]) break;
          // download as early as possible, i.e. no heuristic break
          // tentatively include jjjj:
          ci++; 
          jjj++;
          prevprevjjjj=prevjjjj; 
          prevjjjj=jjjj; 
       }
       //printf("            ... ci=%d\n", ci ); 
       if (i<N-1 && down_possible[i+1][jj]) {
        // find cipp = # vectors jjjj following jj in memory 
        //     for which down_possible is also 1 at time i+1
        jjj=j+1;
        prevjjjj = jj;
        cipp=1;
        while (jjj<k) { 
          int jjjj = rev_position[jjj];
          if (!down_possible[i+1][jjjj]) break;
          if (downloaded[jjjj]) break;
          // also, break if there is no affinity edge from previous jjjj to this one.
          if (!aff[prevjjjj][jjjj]) break;
          cipp++; 
          jjj++;
          prevjjjj=jjjj; 
        }
        //printf("            ... cipp=%d\n", cipp ); 
       }
       // greedy strategy to find best co-downloads:
       if (i>=N-1 || ci > cipp) { // download pos j...j+ci at time i
          //printf("             ... download from pos %d to %d at time %d\n", j, j+ci-1, i);
          assert(ci>0);
          if (downloads_from[i]==-1)
             downloads_from[i] = j;   // otherwise keep earlier lower bound
          downloads_until[i] = j+ci-1;
          for (jjj=j; jjj<j+ci; jjj++) {
             //printf("           %d -> %c\n", jjj, lett(rev_position[jjj]) );
             downloaded[rev_position[jjj]] = 1; // mark as done for download
          }
       }
    }
 }

 // could later insert some checking code here to make sure that
 // all containers with a consumer on GPU are uploaded on time
 // and ... with a producer on GPU are downloaded on time...

 for (j=0; j<k; j++)  // reset flags for emitting upload/download code:
    uploaded[j] = downloaded[j] = 0; 

 savings = 0; // record number of fused messages

 gettimeofday( &printstarttime, NULL);  // take the time 181119

 for (i=0; i<N; i++) { // emit kernels:
    // bulk uploads before i:
    if (uploads_from[i] > -1) { // bulk uploads to do (first):
       for (j=uploads_from[i]; j<=uploads_until[i]; j++) {
          assert(uploaded[rev_position[j]] == 0);
          printf(" before %d: %s-upload %c from location %d\n", 
                 i, ((uploads_until[i]-uploads_from[i])>1)?"bulk":"single",
                 lett(rev_position[j]), j );
          uploaded[rev_position[j]] = 1;
       }
       savings += uploads_until[i]-uploads_from[i];
       if (fp) {
          fprintf(fp, "  cudaMemcpy( d_%c, h_%c, %d*VLENGTH*sizeof(float), cudaMemcpyHostToDevice );\n",
                  lett(rev_position[uploads_from[i]]),
                  lett(rev_position[uploads_from[i]]),
                  uploads_until[i]-uploads_from[i]+1 );
       }
    }
    if (device[i]==1) { // make sure that all operands of i are there:
      for (j=0; j<indeg[i]; j++) {
        if (in[i][j]<INIT_K || device[prod[in[i][j]][0]]==0) {
          if (!uploaded[in[i][j]]) {
            printf(" before %d: single-upload %c from location %d\n", 
                   i, lett(in[i][j]), position[in[i][j]] );
            uploaded[in[i][j]] = 1;
            if (fp) {
              fprintf(fp, "  cudaMemcpy( d_%c, h_%c, VLENGTH*sizeof(float), cudaMemcpyHostToDevice );\n",
                      lett(rev_position[j]), lett(rev_position[j]) );
            }
          }
        }
      }
    }
#if 0
    // old code, to be revised:
    for (j=0; j<indeg[i]; j++) { // all input operands
       int jj, jjj, jjjj;
       if (!uploaded[in[i][j]]) {
          int vj = in[i][j];
          // vj must be loaded before i. Not necessarily first.
          // now go through other non-uploaded vectors 
          // that overlap in their earliest_up...latest_up interval
          // and are connected to j in bh.
          // finally, emit in position[] order.
          for (jj=position[vj]; jj>=0; jj--) { // search to left in position
            int v = rev_position[jj];
            if (uploaded[v]) break;
            //... if no overlap, break; TODO
          }
          for (jjj=position[vj]; jjj<k; jjj++) { // search to right in position
            int v = rev_position[jj];
            if (uploaded[v]) break;
            //... if no overlap, break; TODO
            // ... TODO ...
          }
          // candidates jj...j...jjj can be uploaded in this order:
          for (jjjj=jj; jjjj<=jjj; jjjj++) { 
            int v = rev_position[jjjj];
            printf("upload %c\n", lett(jjjj)); // test
            uploaded[v] = 1;
          }
       }
    }
    // end old code to be revised or dropped
#endif

    printf("  Kernel %d (%s) ( " /*)*/, i, name[i] );
    if (fp)
       if (device[i] == 0) // CPU:
          fprintf(fp, "  CPU%s_%d_%d ( VLENGTH,", name[i], indeg[i], outdeg[i] );
       else // GPU:
          fprintf(fp, "  %s_%d_%d <<< (VLENGTH+255)/256,256>>>( VLENGTH, ", name[i], indeg[i], outdeg[i] );
    // argument list:
    for (j=0; j<indeg[i]; j++) {
       printf(" R(%c),", lett(in[i][j]));
       if (fp) 
         if (device[i]==0)
           fprintf(fp, " h_%c%c", lett(in[i][j]), (outdeg[i]==0&&j==indeg[i]-1)?' ':',');
         else 
           fprintf(fp, " d_%c%c", lett(in[i][j]), (outdeg[i]==0&&j==indeg[i]-1)?' ':',');
    }
    for (j=0; j<outdeg[i]; j++) {
       printf(" W(%c)%c", lett(out[i][j]), j==outdeg[i]-1?' ':',');
       if (fp)
         if (device[i]==0)
           fprintf(fp, " h_%c%c", lett(out[i][j]), j==outdeg[i]-1?' ':',');
         else
           fprintf(fp, " d_%c%c", lett(out[i][j]), j==outdeg[i]-1?' ':',');
    }
    if (device[i] == 0) {
       printf(" ) on CPU\n");
       if (fp) fprintf( fp, "); /*CPU Kernel call %d*/\n", i );
    }
    else {
       printf(" ) on GPU\n");
       if (fp) fprintf( fp, " ); /*GPU Kernel call %d*/\n", i );
    }

    // bulk-downloads after i:
    if (downloads_from[i] > -1) { // bulk downloads to do (first):
       for (j=downloads_from[i]; j<=downloads_until[i]; j++) {
          if( downloaded[rev_position[j]] != 0 )
            printf("??? i=%d j=%d vj=%c dnl=%d\n", i, j, lett(rev_position[j]),
                     downloaded[rev_position[j]] );
// BUG: FIXME - assert now eliminated to get through as it only occurs for the last kernel call:
          //assert( downloaded[rev_position[j]] == 0 );
          printf(" after %d: bulk-download %c to location %d\n", 
                 i, lett(rev_position[j]), j );
          downloaded[rev_position[j]] = 1;
       }
       savings += downloads_until[i]-downloads_from[i];
       if (fp)
          fprintf( fp, "  cudaMemcpy( h_%c, d_%c, %d*VLENGTH*sizeof(float), cudaMemcpyDeviceToHost );\n",
                   lett(rev_position[downloads_from[i]]),
                   lett(rev_position[downloads_from[i]]),
                   downloads_until[i]-downloads_from[i]+1 );
    }
#if 0 
    // we do not require all produced values to be downloaded immediately... 
    if (device[i]==1) 
    for (j=0; j<outdeg[i]; j++) {
       if (!downloaded[out[i][j]]) {
          printf(" after %d:  download %c to location %d\n", 
                 i, lett(out[i][j]), position[out[i][j]] );
          downloaded[out[i][j]] = 1;
       }
    }
#endif
 } // for i - loop over kernel calls
#if 0
 // it seems the following is not necessary and only leads to trouble:

 // finally, downloads after N-1 for live on exit vectors:
 if (downloads_from[N] > -1) { // bulk downloads to do (first):
   for (j=downloads_from[N]; j<=downloads_until[N]; j++) {
      printf("HALLO! j=%d\n", j);
      if( downloaded[rev_position[j]] != 0 )
        printf("??? i=N=%d j=%d vj=%c dnl=%d\n", N, j, lett(rev_position[j]),
                 downloaded[rev_position[j]] );
      assert( downloaded[rev_position[j]] == 0 );
      printf(" after %d: download %c to location %d\n", 
             N, lett(rev_position[j]), j );
      downloaded[rev_position[j]] = 1;
   }
   savings += downloads_until[N]-downloads_from[N];
 }
#endif
   // pending downloads of live-on-exit vectors:
   for (j=INIT_K; j<k; j++) {
     if //(down_possible[N][j] && !downloaded[j]) 
        // Fixed bug 190411: need download only if any consumer is on CPU:
        (device[prod[j][0]]==1 && !downloaded[j]
	        && (   device[cons[j][0]]==0
		       || ncons[j]>=2 && device[cons[j][1]]==0
		       || ncons[j]>=3 && device[cons[j][2]]==0
		       || ncons[j]>=4 && device[cons[j][3]]==0
		       || ncons[j]>=5 && device[cons[j][4]]==0
		       || ncons[j]>=6 && device[cons[j][5]]==0
		       || ncons[j]>=7 && device[cons[j][6]]==0
		       || ncons[j]>=8 && device[cons[j][7]]==0  // enough...
		   ) 
		)
     {
        printf(" after %d:   download %c to location %d\n", N, lett(j), position[j] );
        downloaded[j] = 1;
        if (fp) fprintf( fp, "  cudaMemcpy( h_%c, d_%c, VLENGTH*sizeof(float), cudaMemcpyDeviceToHost );\n",
                         lett(j), lett(j) );
     }
   }
   // no savings here
   
   // Now deallocate allocated vectors / vector groups:
   if (fp) {
     for (j=0; j<k; j++) {
        if (calledmalloc[j])
	   fprintf(fp, "  free( h_%c );\n  cudaFree( d_%c );\n", lett(j), lett(j) ); 
     }
   }

   if (fp) { // epilogue, including timing loop:
      fprintf(fp, " }\n gettimeofday( &fendtime, NULL);\n fusedtime=(fendtime.tv_sec-fstarttime.tv_sec)*1000000 + (fendtime.tv_usec-fstarttime.tv_usec);\n printf(\" Time (avgd. over %%d iter.) with transfer fusion: %%d us\\n\", NITER, fusedtime/NITER);\n}\n");
   }
}



void main( int argc, char *argv[] )
{
 char s[256];
 int opttime, livetime, eltime, afftime, hamtime, emittime, printtime;
 FILE *cudafp, *ncudafp, *acudafp, *ancudafp;
 char cudafilename[256], ncudafilename[256], acudafilename[256], ancudafilename[256];

 sprintf( ncudafilename, "cuda/n%s.cu", argv[0]+2);
 printf("Naive CUDA source code output to file %s\n", ncudafilename );
 if ((ncudafp = fopen( ncudafilename, "w" )) == NULL) {
    printf("Cannot open file %s\n", ncudafilename+2);
    exit(1);
 }
 sprintf( ancudafilename, "cuda/an%s.cu", argv[0]+2);
 printf("Naive CUDA source code output incl. allocation measurement to file %s\n", ancudafilename );
 if ((ancudafp = fopen( ancudafilename, "w" )) == NULL) {
    printf("Cannot open file %s\n", ancudafilename+2);
    exit(1);
 }
 sprintf( cudafilename, "cuda/%s.cu", argv[0]+2);
 printf("Transfer-fused CUDA source code output to file %s\n", cudafilename );
 if ((cudafp = fopen( cudafilename, "w" )) == NULL) {
    printf("Cannot open file %s\n", cudafilename+2);
    exit(1);
 }
 sprintf( acudafilename, "cuda/a%s.cu", argv[0]+2);
 printf("Transfer-fused CUDA source code incl. allocation measurement to file %s\n", acudafilename );
 if ((acudafp = fopen( acudafilename, "w" )) == NULL) {
    printf("Cannot open file %s\n", acudafilename+2);
    exit(1);
 }

 build_DAG();

 // create dot file for visualization:
 //sprintf( s, "dot/%s_N%d_k%d_Ind%d_%d_Outd%d_%d.dot",
 //           argv[0], N, k, MIN_INDEG, MAX_INDEG, MIN_OUTDEG, MAX_OUTDEG);
 // writedot_DAG( s );  // trouble in dot/makefile
 writedot_DAG( "blubber.dot" );
 
 // calculate_schedule(); // later 

 gettimeofday( &starttime, NULL);
 calculate_live();

#ifdef SAMECOLOR
 samecolor();
#else
 linearscan();  // more memory efficient but more constraining for affs
#endif

 gettimeofday( &elstarttime, NULL);
 earliestlatest(); // calculate up_earliest, up_latest, down_earliest, down_latest

 gettimeofday( &affstarttime, NULL);
 calculate_aff();
 
 gettimeofday( &hamstarttime, NULL);
 comp_hamiltonians();

 // writedot_Aff

 gettimeofday( &emitnaivestarttime, NULL);

 // for timing,  pass NULL to suppress source code generation
 // emit_code( NULL );  // the outcome of the optimization, calc. savings
 emit_naive( ncudafp );  // the outcome of the optimization, calc. savings
 //gettimeofday( &emitnaiveendtime, NULL);
 emit_naive_alloc( ancudafp );  // the outcome of the optimization, calc. savings incl alloc fusion

 emit_code( cudafp );  // the outcome of the optimization, calc. savings
 // put it here to count the emit time only once:
 gettimeofday( &emitstarttime, NULL);
 emit_code_alloc( acudafp );  // the outcome of the optimization, calc. savings incl alloc fusion

 gettimeofday( &endtime, NULL);
 // We do not include the code print time (100-200us) in the optimization time,
 // and neither the time for calculating live[] and linearscan which we do not use:
 //opttime = (printstarttime.tv_sec - elstarttime.tv_sec) * 1000000
 //           + (printstarttime.tv_usec - elstarttime.tv_usec);
 livetime = (elstarttime.tv_sec - starttime.tv_sec) * 1000000
            + (elstarttime.tv_usec - starttime.tv_usec);
 eltime = (affstarttime.tv_sec - elstarttime.tv_sec) * 1000000
            + (affstarttime.tv_usec - elstarttime.tv_usec);
 afftime = (hamstarttime.tv_sec - affstarttime.tv_sec) * 1000000
            + (hamstarttime.tv_usec - affstarttime.tv_usec);
 hamtime = (emitnaivestarttime.tv_sec - hamstarttime.tv_sec) * 1000000
            + (emitnaivestarttime.tv_usec - hamstarttime.tv_usec);
 emittime = (printstarttime.tv_sec - emitstarttime.tv_sec) * 1000000
            + (printstarttime.tv_usec - emitstarttime.tv_usec);
 printtime = (endtime.tv_sec - printstarttime.tv_sec) * 1000000
            + (endtime.tv_usec - printstarttime.tv_usec);

 writedot_Aff( "aff.dot" );  // plot Aff incl. best Hamiltonian

 // calculate opttime from subparts as emitting for the 3 other variants should not be counted:
 opttime = eltime + afftime + hamtime + emittime;
 printf("Savings: %d startups.    Time: %d microseconds (EarLat: %d, Aff: %d, Ham: %d, Emit: %d). Live: %d, Print: %d\n", savings, opttime, eltime, afftime, hamtime, emittime, livetime, printtime );
}


